import math
import logging
from os import path
from typing import List, Union

from keras import layers, models
import numpy as np
import cv2

from .. import core, utils
from .detector import Detector

log = logging.getLogger(__name__)

""" MTCNN
This intends to replicate the MTCNN architecture that comes with FaceNet. It
has one small change from the original (see detect function). To compare
outputs, you can use something like the following.

from mira import core, detectors
from mtcnn.mtcnn import MTCNN
import matplotlib.pyplot as plt

url = 'https://www.abc.net.au/news/image/9499864-3x2-700x467.jpg'
image = core.Image.read(url)

detector1 = detectors.MTCNN()
detector2 = MTCNN()

faces1 = detector1.detect(image)
faces2 = detector2.detect_faces(image)
scene1 = core.Scene(
    annotations=faces1,
    image=image,
    annotation_config=detector1.annotation_config
)
scene2 = core.Scene(
    annotation_config=detector1.annotation_config,
    annotations=[core.Annotation(
        category=detector1.annotation_config['face'],
        selection=core.Selection([
            [f['box'][0], f['box'][1]],
            [f['box'][0] + f['box'][2], f['box'][1]+f['box'][3]]
        ])
    ) for f in faces2],
    image=image
)
print('Our boxes:', [f.selection.xywh() for f in faces1])
print('Their boxes:', [f['box'] for f in faces2])
fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(10, 10))
scene1.show(ax1)
scene2.show(ax2)
"""

WEIGHTS_CONFIG = {
    "pnet": {
        "hash":
        "3285cf7a3de2651c5784cb9e32013f5919aae95fd1ed1bc371dd9691affb39af",  # noqa: E501
        "url":
        "https://storage.googleapis.com/miradata/weights/mtcnn/det1.npy",
    },
    "rnet": {
        "hash":
        "716b8b83e42476791c9096f14dbb09fefc88bf5c7ec876b1683f9acd52b3f39c",  # noqa: E501
        "url":
        "https://storage.googleapis.com/miradata/weights/mtcnn/det2.npy",
    },
    "onet": {
        "hash":
        "396ead803d85d3443307ff8f45fb6aed2536579b415a4f4d4cb8f93ea6b1476a",  # noqa: E501
        "url":
        "https://storage.googleapis.com/miradata/weights/mtcnn/det3.npy",
    },
}


def make_pnet(input_shape=(None, None, 3)):
    x = layers.Input(input_shape)
    conv1 = layers.Conv2D(kernel_size=(3, 3),
                          filters=10,
                          strides=(1, 1),
                          padding="valid",
                          name="conv1")(x)
    relu1 = layers.PReLU(name="PReLU1", shared_axes=[1, 2])(conv1)
    mp1 = layers.MaxPooling2D(pool_size=(2, 2),
                              strides=(2, 2),
                              padding="same",
                              name="pool1")(relu1)
    conv2 = layers.Conv2D(kernel_size=(3, 3),
                          filters=16,
                          strides=(1, 1),
                          padding="valid",
                          name="conv2")(mp1)
    relu2 = layers.PReLU(name="PReLU2", shared_axes=[1, 2])(conv2)
    conv3 = layers.Conv2D(kernel_size=(3, 3),
                          filters=32,
                          strides=(1, 1),
                          padding="valid",
                          name="conv3")(relu2)
    relu3 = layers.PReLU(name="PReLU3", shared_axes=[1, 2])(conv3)
    conv4_1 = layers.Conv2D(kernel_size=(1, 1),
                            filters=2,
                            strides=(1, 1),
                            padding="same",
                            name="conv4-1")(relu3)
    prob1 = layers.Softmax(axis=3, name="prob1")(conv4_1)
    conv4_2 = layers.Conv2D(kernel_size=(1, 1),
                            filters=4,
                            strides=(1, 1),
                            padding='same',
                            name="conv4-2")(relu3)
    return models.Model(inputs=x, outputs=[prob1, conv4_2])


def make_rnet(input_shape=(24, 24, 3)):
    x = layers.Input(input_shape)
    conv1 = layers.Conv2D(kernel_size=(3, 3),
                          filters=28,
                          strides=(1, 1),
                          padding="valid",
                          name="conv1")(x)
    relu1 = layers.PReLU(name="prelu1", shared_axes=[1, 2])(conv1)
    mp1 = layers.MaxPooling2D(pool_size=(3, 3),
                              strides=(2, 2),
                              padding="same",
                              name="pool1")(relu1)
    conv2 = layers.Conv2D(kernel_size=(3, 3),
                          filters=48,
                          strides=(1, 1),
                          padding="valid",
                          name="conv2")(mp1)
    relu2 = layers.PReLU(name="prelu2", shared_axes=[1, 2])(conv2)
    mp2 = layers.MaxPooling2D(pool_size=(3, 3),
                              strides=(2, 2),
                              padding="valid",
                              name="pool2")(relu2)
    conv3 = layers.Conv2D(kernel_size=(2, 2),
                          filters=64,
                          strides=(1, 1),
                          padding="valid",
                          name="conv3")(mp2)
    relu3 = layers.PReLU(name="prelu3", shared_axes=[1, 2])(conv3)
    conv4 = layers.Conv2D(kernel_size=(3, 3),
                          filters=128,
                          strides=(1, 1),
                          padding="valid",
                          name="conv4")(relu3)
    relu4 = layers.PReLU(name="prelu4", shared_axes=[1, 2])(conv4)
    conv5_1 = layers.Conv2D(kernel_size=(1, 1),
                            filters=2,
                            strides=(1, 1),
                            padding="valid",
                            name="conv5-1")(relu4)
    prob1 = layers.Softmax(axis=3)(conv5_1)
    conv5_2 = layers.Conv2D(kernel_size=(1, 1),
                            filters=4,
                            strides=(1, 1),
                            padding="valid",
                            name="conv5-2")(relu4)
    clf = layers.Reshape((2, ), name='clf')(prob1)
    reg = layers.Reshape((4, ), name='reg')(conv5_2)
    return models.Model(inputs=x, outputs=[clf, reg])


def make_onet(input_shape=(48, 48, 3)):
    x = layers.Input(input_shape)
    conv1 = layers.Conv2D(kernel_size=(3, 3),
                          filters=32,
                          strides=(1, 1),
                          padding="valid",
                          name="conv1")(x)
    relu1 = layers.PReLU(name="prelu1", shared_axes=[1, 2])(conv1)
    mp1 = layers.MaxPooling2D(pool_size=(3, 3),
                              strides=(2, 2),
                              padding="same",
                              name="pool1")(relu1)
    conv2 = layers.Conv2D(kernel_size=(3, 3),
                          filters=64,
                          strides=(1, 1),
                          padding="valid",
                          name="conv2")(mp1)
    relu2 = layers.PReLU(name="prelu2", shared_axes=[1, 2])(conv2)
    mp2 = layers.MaxPooling2D(pool_size=(3, 3),
                              strides=(2, 2),
                              padding="valid",
                              name="pool2")(relu2)
    conv3 = layers.Conv2D(kernel_size=(3, 3),
                          filters=64,
                          strides=(1, 1),
                          padding="valid",
                          name="conv3")(mp2)
    relu3 = layers.PReLU(name="prelu3", shared_axes=[1, 2])(conv3)
    mp3 = layers.MaxPooling2D(pool_size=(2, 2),
                              strides=(2, 2),
                              padding="same",
                              name="pool3")(relu3)
    conv4 = layers.Conv2D(kernel_size=(2, 2),
                          filters=128,
                          strides=(1, 1),
                          padding="valid",
                          name="conv4")(mp3)
    relu4 = layers.PReLU(name="prelu4", shared_axes=[1, 2])(conv4)
    conv5 = layers.Conv2D(kernel_size=(3, 3),
                          filters=256,
                          padding="valid",
                          name="conv5")(relu4)
    relu5 = layers.PReLU(name="prelu5", shared_axes=[1, 2])(conv5)
    conv6_1 = layers.Conv2D(kernel_size=(1, 1), filters=2,
                            name="conv6-1")(relu5)
    prob1 = layers.Softmax(axis=3, name="prob1")(conv6_1)
    conv6_2 = layers.Conv2D(kernel_size=(1, 1), filters=4,
                            name="conv6-2")(relu5)
    conv6_3 = layers.Conv2D(kernel_size=(1, 1), filters=10,
                            name="conv6-3")(relu5)
    clf = layers.Reshape((2, ), name='clf')(prob1)
    reg = layers.Reshape((4, ), name='reg')(conv6_2)
    feats = layers.Reshape((10, ), name='feats')(conv6_3)
    return models.Model(inputs=x, outputs=[clf, reg, feats])


def load_weights(network, weights_file):
    for layer_name, weights in np.load(
        weights_file,
        encoding="latin1",
        allow_pickle=True
    ).item().items():
        layer = network.get_layer(layer_name)
        if layer_name.lower().startswith("conv"):
            coefs = weights["weights"].reshape(
                layer.weights[0].shape.as_list())
            biases = weights["biases"].reshape(
                layer.weights[1].shape.as_list())
            layer.set_weights(weights=[coefs, biases])
        elif layer_name.lower().startswith("prelu"):
            alpha = weights["alpha"].reshape(layer.weights[0].shape)
            layer.set_weights(weights=[alpha])
        else:
            raise NotImplementedError("Cannot handle layer: " + layer_name)


def nms(boxes, threshold, method):
    if boxes.size == 0:
        return np.empty((0, 9))

    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2]
    y2 = boxes[:, 3]
    s = boxes[:, 4]

    area = (x2 - x1 + 1) * (y2 - y1 + 1)
    sorted_s = np.argsort(s)

    pick = np.zeros_like(s, dtype=np.int16)
    counter = 0
    while sorted_s.size > 0:
        i = sorted_s[-1]
        pick[counter] = i
        counter += 1
        idx = sorted_s[0:-1]

        xx1 = np.maximum(x1[i], x1[idx])
        yy1 = np.maximum(y1[i], y1[idx])
        xx2 = np.minimum(x2[i], x2[idx])
        yy2 = np.minimum(y2[i], y2[idx])

        w = np.maximum(0.0, xx2 - xx1 + 1)
        h = np.maximum(0.0, yy2 - yy1 + 1)

        inter = w * h

        if method == 'Min':
            o = inter / np.minimum(area[i], area[idx])
        else:
            o = inter / (area[i] + area[idx] - inter)

        sorted_s = sorted_s[np.where(o <= threshold)]

    pick = pick[0:counter]

    return boxes[pick]


def padded(x1y1x2y2s):
    x1y1x2y2s = x1y1x2y2s.copy()
    x1y1x2y2s[:, :2] = np.clip(x1y1x2y2s[:, :2], a_min=1, a_max=np.inf) - 1
    return x1y1x2y2s


def square(x1y1x2y2s):
    x1y1x2y2s = x1y1x2y2s.copy()
    h = x1y1x2y2s[:, 3] - x1y1x2y2s[:, 1]
    w = x1y1x2y2s[:, 2] - x1y1x2y2s[:, 0]
    length = np.maximum(w, h)
    x1y1x2y2s[:, 0] = x1y1x2y2s[:, 0] + w * 0.5 - length * 0.5
    x1y1x2y2s[:, 1] = x1y1x2y2s[:, 1] + h * 0.5 - length * 0.5
    x1y1x2y2s[:, 2:4] = x1y1x2y2s[:, 0:2] + length[:, np.newaxis]
    x1y1x2y2s[:, :4] = np.fix(x1y1x2y2s[:, :4])
    return x1y1x2y2s


def regress(boxes):
    boxes = boxes.copy()
    w = boxes[:, 2] - boxes[:, 0]
    h = boxes[:, 3] - boxes[:, 1]
    boxes[:, 0] += boxes[:, 5] * w
    boxes[:, 1] += boxes[:, 6] * h
    boxes[:, 2] += boxes[:, 7] * w
    boxes[:, 3] += boxes[:, 8] * h
    boxes[:, 5:] = 0
    w = boxes[:, 2] - boxes[:, 0]
    h = boxes[:, 3] - boxes[:, 1]
    return boxes


def fix_scale(x):
    return (x.astype('float64') - 127.5) * 0.0078125


def run_stage(network, image, x1y1x2y2s, cell_size):
    # we need to use a padded version of x1y1x2y2s, I guess?
    if len(x1y1x2y2s) == 0:
        return np.empty((0, 9))
    crops = [
        image[y1:y2, x1:x2].astype('float64')
        for x1, y1, x2, y2 in padded(x1y1x2y2s[:, :4]).astype('int32')
    ]
    crops = [
        image for image in crops if image.shape[0] > 0 and image.shape[1] > 0
    ]
    crops = [
        cv2.resize(image, (cell_size, cell_size), interpolation=cv2.INTER_AREA)
        for image in crops
    ]
    X = np.float64(crops)
    X = fix_scale(X.transpose(0, 2, 1, 3))
    clf, reg = [v for v in network.predict(X)[:2]]

    x1y1x2y2s[:, 4] = clf[:, 1]
    x1y1x2y2s[:, :4] = np.fix(x1y1x2y2s[:, :4])
    x1y1x2y2s[:, 5:] = reg
    return x1y1x2y2s


class MTCNN(Detector):
    """A detector wrapping MTCNN for face detection. Based on the
    implementation found as part of
    `facenet <https://github.com/davidsandberg/facenet>`_. The only
    difference is we've implemented it using keras and we make one
    tweak to when nms is applied for Stage 2. See comments in
    code for detail. Training is not supported at this time. We also
    do not support recording the locations of facial landmarks. The model
    is documented in `the original paper
    <https://kpzhang93.github.io/MTCNN_face_detection_alignment/paper/spl.pdf>`_.

    Example usage:

    .. highlight:: python
    .. code-block:: python

        from mira import core, detectors

        url = 'https://www.abc.net.au/news/image/9499864-3x2-700x467.jpg'
        image = core.Image.read(url)
        detector = detectors.MTCNN()
        faces = detector.detect(image)
        scene = core.Scene(
            annotations=faces,
            image=image,
            annotation_config=detector.annotation_config
        )

    Attributes:
        pnet: The pnet model (stage 1)
        rnet: The rnet model (stage 2)
        onet: The onet model (stage 3)
    """
    def __init__(self):
        self.annotation_config = core.AnnotationConfiguration(["face"])
        self.pnet = make_pnet()
        self.rnet = make_rnet()
        self.onet = make_onet()

        load_weights(
            self.pnet,
            utils.get_file(
                origin=WEIGHTS_CONFIG["pnet"]["url"],  # noqa; E501
                file_hash=WEIGHTS_CONFIG["pnet"]["hash"],  # noqa; E501
                cache_subdir=path.join("weights", "mtcnn"),
                hash_algorithm="sha256",
                extract=False,
            ),
        )
        load_weights(
            self.rnet,
            utils.get_file(
                origin=WEIGHTS_CONFIG["rnet"]["url"],  # noqa; E501
                file_hash=WEIGHTS_CONFIG["rnet"]["hash"],  # noqa; E501
                cache_subdir=path.join("weights", "mtcnn"),
                hash_algorithm="sha256",
                extract=False,
            ),
        )
        load_weights(
            self.onet,
            utils.get_file(
                origin=WEIGHTS_CONFIG["onet"]["url"],  # noqa; E501
                file_hash=WEIGHTS_CONFIG["onet"]["hash"],  # noqa; E501
                cache_subdir=path.join("weights", "mtcnn"),
                hash_algorithm="sha256",
                extract=False,
            ),
        )

    def compute_inputs(self, images):
        raise NotImplementedError('MTCNN is not yet complete.')

    def compute_targets(self, collection):
        raise NotImplementedError('MTCNN is not yet complete.')

    def detect(
        self,
        image: Union[core.Image, np.ndarray],
        minsize=20,
        factor=0.709,
        thresholds=None
    ) -> List[core.Annotation]:
        """Detects faces in an image, and returns bounding boxes and points for them.

        image: input image
        minsize: Minimum face size
        factor: The factor by which to step down scales in the
         image.
        thresholds: The detection thresholds (one for each of three stages)
        """
        if thresholds is None:
            thresholds = [0.6, 0.7, 0.7]
        image = image.view(core.Image)
        height, width = image.shape[:2]

        # Scale of MTCNN cell to minimum face size
        m = 12.0 / minsize

        # Reequired image size to have a 12x12
        # region correspond to minimum face
        # size
        minl = min([height, width]) * m
        n_scales = math.ceil(math.log(12 / minl, factor))
        scales = [m * np.power(factor, n) for n in range(n_scales)]

        # Stage 1
        x1y1x2y2s = []
        stride = 2
        cell_size = 12
        for idx, scale in enumerate(scales):
            scaled_image = image.resize(scale, interpolation=cv2.INTER_AREA)
            X = np.float64([
                fix_scale(scaled_image.transpose(1, 0, 2))
            ])
            y = [v.transpose(0, 2, 1, 3) for v in self.pnet.predict(X)]
            clf, reg = y[0][0], y[1][0]
            yc, xc = np.where(clf[..., 1].transpose() >= thresholds[0])
            score = clf[xc, yc, 1:]
            reg = reg[xc, yc]
            bb = np.vstack([yc, xc]).transpose()
            x1y1 = np.fix((stride * bb + 1) / scale)
            x2y2 = np.fix((stride * bb + 1 + cell_size - 1) / scale)
            x1y1x2y2s_current = np.concatenate(
                [x1y1, x2y2, score, reg], axis=1
            )
            x1y1x2y2s.append(nms(x1y1x2y2s_current, 0.5, "Union"))
        x1y1x2y2s = np.concatenate(x1y1x2y2s)
        x1y1x2y2s = nms(x1y1x2y2s, 0.7, "Union")
        x1y1x2y2s = regress(x1y1x2y2s)
        x1y1x2y2s = square(x1y1x2y2s)

        # Stage 2
        x1y1x2y2s = run_stage(self.rnet, image, x1y1x2y2s, 24)
        x1y1x2y2s = x1y1x2y2s[x1y1x2y2s[:, 4] > thresholds[1]]

        # Note the ordering here of applying regression and nms differs from
        # the original implementation. In the original, nms is applied first
        # and regression second on Stage 2. But before regression, the boxes
        # are the same as they were before, we just have a smaller subset of
        # them, making the nms call superfluous. I've swapped the order here
        # to be consistent with Stage 3, where nms is applied after regression
        # regression. To recover the original behavior, the nms and regression
        # lines should be swapped back.
        # https://github.com/davidsandberg/facenet/blob/096ed770f163957c1e56efa7feeb194773920f6e/src/align/detect_face.py#L377
        x1y1x2y2s = regress(x1y1x2y2s)
        x1y1x2y2s = nms(x1y1x2y2s, 0.7, 'Union')
        x1y1x2y2s = square(x1y1x2y2s)

        # Stage 3
        x1y1x2y2s = run_stage(self.onet, image, x1y1x2y2s, 48)
        x1y1x2y2s = x1y1x2y2s[x1y1x2y2s[:, 4] > thresholds[2]]
        x1y1x2y2s = regress(x1y1x2y2s)
        x1y1x2y2s = nms(x1y1x2y2s, 0.7, 'Min')

        return [
            core.Annotation(
                selection=core.Selection(points=[[x1, y1], [x2, y2]]),
                category=self.annotation_config['face'],
                score=s
            ) for x1, y1, x2, y2, s in x1y1x2y2s[:, :5]
        ]



import logging
from typing import List
from pkg_resources import resource_string

import numpy as np

from .selection import Selection

log = logging.getLogger(__name__)


class AnnotationCategory:
    """Defines a category of an annotation along
    with all associated properties.

    Args:
        name: The name of the annotation category

    """
    def __init__(self, name: str):
        self._name = name

    def __eq__(self, other):
        return self._name == other._name

    @property
    def name(self):
        return self._name


class AnnotationConfiguration:
    """A class defining a list of annotation
    types for an object detection class.

    Args:
        names: The list of class names
    """
    def __init__(self, names: List[str]):
        names = [s.lower() for s in names]
        if len(names) != len(set(names)):
            raise ValueError(
                'All class names must be unique '
                '(case-insensitive).'
            )
        self._types = [
            AnnotationCategory(
                name=name
            ) for name in names
        ]

    def __getitem__(self, key):
        if type(key) == np.int64:
            key = int(key)
        if type(key) == int:
            if key >= len(self):
                raise ValueError(
                    'Index {0} is out of bounds '.format(key) +
                    '(only have {0} entries)'.format(len(self))
                )
            return self.types[key]
        elif type(key) == str:
            key = key.lower()
            val = next((e for e in self._types if e.name == key), None)
            if val is None:
                raise ValueError(
                    'Did not find {0} in configuration'.format(key)
                )
            return val
        else:
            raise ValueError(
                'Key must be int or str, not ' + str(type(key))
            )

    def __iter__(self):
        return iter(self._types)

    def __contains__(self, key):
        if type(key) == str:
            return any(e.name == key for e in self._types)
        elif type(key) == AnnotationCategory:
            return any(e == key for e in self._types)
        else:
            raise ValueError(
                'Key must be str or AnnotationCategory, not '
                '' + str(type(key))
            )

    def __eq__(self, other):
        if type(other) != type(self):
            return False
        if len(other) != len(self):
            return False
        return all(o == s for s, o in zip(self, other))

    def __len__(self):
        return len(self._types)

    @property
    def types(self):
        return self._types

    def index(self, category):
        return next(i for i, cat in enumerate(self) if cat == category)


AnnotationConfiguration.COCO = AnnotationConfiguration(
    resource_string(
        __name__, 'assets/coco_classes.txt'
    ).decode('utf-8').split('\n')
)
AnnotationConfiguration.VOC = AnnotationConfiguration(
    resource_string(
        __name__, 'assets/voc_classes.txt'
    ).decode('utf-8').split('\n')
)


class Annotation:
    """Defines a single annotation.

    Args:
        selection: The selection associated with the annotation
        category: The category of the annotation
        score: A score for the annotation
    """
    def __init__(
        self,
        selection: Selection,
        category: AnnotationCategory,
        score: float=None
    ):
        if category is None:
            raise ValueError(
                'A category object must be specified.'
            )
        self.selection = selection
        self.category = category
        self.score = score

    def assign(self, **kwargs) -> 'Annotation':
        """Get a new Annotation with only the supplied
        keyword arguments changed."""
        defaults = {
            'selection': self.selection,
            'category': self.category,
            'score': self.score
        }
        kwargs = {**defaults, **kwargs}
        return Annotation(**kwargs)

    def convert(self, annotation_config) -> 'Annotation':
        name = self.category.name
        if name in annotation_config:
            return self.assign(
                category=annotation_config[name],
            )
        else:
            log.warning(
                '{0} is not in the new annotation '
                'configuration.'.format(name)
            )
            return None

    def resize(self, scale) -> 'Annotation':
        """Get a new annotation with the given
        uniform scaling."""
        return self.assign(selection=self.selection.resize(scale=scale))

    def __eq__(self, other):
        self_bbox = self.selection.bbox()
        other_bbox = other.selection.bbox()
        return (
            self_bbox == other_bbox and
            self.category == other.category
        )


import logging
from os import path
from itertools import product

from glob import glob
import numpy as np
import cv2
import re

from .. import utils
from ..core import (
    Image,
    Scene,
    Selection,
    SceneCollection,
    Annotation,
    AnnotationConfiguration
)
from .voc import load_voc


log = logging.getLogger(__name__)


def load_voc2012(subset='train') -> SceneCollection:
    """PASCAL VOC 2012

    Args:
        subset: One of `train`, `val`, or `trainval`.
            If `trainval`, the scene collection contains both the
            train and validation sets. If `train`, only the
            training set. If `val` only the validation
            set.

    Returns:
        A scene collection containing the PASCAL VOC 2012
        dataset.
    """
    assert subset in ['train', 'val', 'trainval'], \
        'Subset must be one of train, val, or trainval'

    def extract_check_fn(root_dir):
        image_dir = path.join(root_dir, 'vocdevkit', 'voc2012', 'jpegimages')
        image_files = glob(path.join(image_dir, '*.jpg'))
        annotation_dir = path.join(root_dir, 'vocdevkit', 'voc2012', 'annotations')  # noqa: E501
        annotation_files = glob(path.join(annotation_dir, '*.xml'))
        return (
            len(annotation_files) == 17125 and
            len(image_files) == 17125
        )
    root_dir = utils.get_file(
        origin='http://pjreddie.com/media/files/VOCtrainval_11-May-2012.tar',  # noqa: E501
        file_hash='e14f763270cf193d0b5f74b169f44157a4b0c6efa708f4dd0ff78ee691763bcb',  # noqa: E501
        cache_subdir=path.join('datasets', 'voc2012'),
        hash_algorithm='sha256',
        extract=True,
        archive_format='tar',
        extract_check_fn=extract_check_fn
    )
    image_dir = path.join(root_dir, 'vocdevkit', 'voc2012', 'jpegimages')
    annotation_dir = path.join(root_dir, 'vocdevkit', 'voc2012', 'annotations')
    imageset_dir = path.join(root_dir, 'vocdevkit', 'voc2012', 'imagesets', 'main')  # noqa: E501
    sids = []
    if subset in ['trainval', 'train']:
        sid_path = path.join(imageset_dir, 'train.txt')
        with open(sid_path, 'r') as f:
            sids.extend(f.read().split('\n'))
    if subset in ['trainval', 'val']:
        sid_path = path.join(imageset_dir, 'val.txt')
        with open(sid_path, 'r') as f:
            sids.extend(f.read().split('\n'))
    filepaths = [
        path.join(annotation_dir, '{0}.xml'.format(sid))
        for sid in sids if len(sid.strip()) > 0
    ]
    return load_voc(
        filepaths=filepaths,
        annotation_config=AnnotationConfiguration.VOC,
        image_dir=image_dir
    )


def load_shapes(
    width=256,
    height=256,
    object_count_bounds=(3, 8),
    object_width_bounds=(20, 40),
    n_scenes=100,
) -> SceneCollection:
    """A simple dataset for testing.

    Args:
        width: The width of each image
        height: The height of each image
        object_count_bounds: A tuple indicating the minimum and maximum
            number of objects in each image
        object_width_bounds: A tuple indicating the minimum and maximum
            widths of the objects in each image
        n_scenes: The number of scenes to generate

    Returns:
        A scene collection of images with circles and rectangles in it.
    """
    annotation_config = AnnotationConfiguration(
        [
            ' '.join([s, c]) for s, c in product(
                ['RED', 'BLUE', 'GREEN'],
                ['RECTANGLE', 'CIRCLE']
            )
        ]
    )

    def make_scene():
        image = Image.new(width=width, height=height, channels=3, cval=255)
        object_count = np.random.randint(*object_count_bounds)
        ws = np.random.randint(*object_width_bounds, size=object_count)
        xs = np.random.randint(
            low=0,
            high=width - object_width_bounds[-1],
            size=object_count
        )
        ys = np.random.randint(
            low=0,
            high=height - object_width_bounds[-1],
            size=object_count
        )
        shapes = np.random.choice(['RECTANGLE', 'CIRCLE'], size=object_count)
        colors = np.random.choice(['RED', 'BLUE', 'GREEN'], size=object_count)
        lookup = {
            'RED': (255, 0, 0),
            'BLUE': (0, 0, 255),
            'GREEN': (0, 255, 0)
        }
        annotations = []
        for x, y, w, shape, color in zip(xs, ys, ws, shapes, colors):
            if shape == 'RECTANGLE':
                cv2.rectangle(
                    image,
                    pt1=(x, y),
                    pt2=(x+w, y+w),
                    thickness=-1,
                    color=lookup[color]
                )
            elif shape == 'CIRCLE':
                r = w // 2
                w = 2*r
                cv2.circle(
                    image,
                    center=(x+r, y+r),
                    radius=r,
                    thickness=-1,
                    color=lookup[color]
                )
            annotations.append(Annotation(
                selection=Selection([
                    [x, y],
                    [x + w, y + w]
                ]),
                category=annotation_config[' '.join([color, shape])]
            ))
        return Scene(
            annotations=annotations,
            image=image,
            annotation_config=annotation_config
        )
    scenes = [make_scene() for n in range(n_scenes)]
    return SceneCollection(
        scenes=scenes,
        annotation_config=annotation_config
    )


def load_oxfordiiitpets(breed=True) -> SceneCollection:
    """Load the Oxford-IIIT pets dataset. It is not divided into
    train, validation, and test because it appeared some files were missing
    from the trainval and test set documents
    (e.g., english_cocker_spaniel_164).

    Args:
        breed: Whether to use the breeds as the class labels. If False, the
            class labels are limited to dog or cat.

    Returns:
        A scene collection containing the dataset
    """
    image_dir = utils.get_file(
        origin='http://www.robots.ox.ac.uk/~vgg/data/pets/data/images.tar.gz',  # noqa: E501
        file_hash='67195c5e1c01f1ab5f9b6a5d22b8c27a580d896ece458917e61d459337fa318d',  # noqa: E501
        cache_subdir=path.join('datasets', 'oxfordiiitpets'),
        hash_algorithm='sha256',
        extract=True,
        archive_format='tar',
        extract_check_fn=lambda directory: len(glob(path.join(directory, 'images', '*.jpg'))) == 7390  # noqa: E501
    )
    annotations_dir = utils.get_file(
        origin='http://www.robots.ox.ac.uk/~vgg/data/pets/data/annotations.tar.gz',  # noqa: E501
        file_hash='52425fb6de5c424942b7626b428656fcbd798db970a937df61750c0f1d358e91',  # noqa: E501
        cache_subdir=path.join('datasets', 'oxfordiiitpets'),
        hash_algorithm='sha256',
        extract=True,
        archive_format='tar',
        extract_check_fn=lambda directory: len(glob(path.join(directory, 'annotations', 'xmls', '*.xml'))) == 3686  # noqa: E501
    )
    filepaths = glob(
        path.join(annotations_dir, 'annotations', 'xmls', '*.xml')
    )
    image_dir = path.join(image_dir, 'images')
    collection = load_voc(
        filepaths=filepaths,
        annotation_config=AnnotationConfiguration(['dog', 'cat']),
        image_dir=image_dir
    )
    if not breed:
        return collection
    assert all(len(s.annotations) in [1, 2] for s in collection.scenes), \
        'An error occurred handling pets dataset'
    labels = ['_'.join(path.splitext(path.split(f)[1])[0].split('_')[:-1]) for f in filepaths]  # noqa: E501
    annotation_config = AnnotationConfiguration(sorted(set(labels)))
    return SceneCollection(
        scenes=[
            scene.assign(
                annotations=[a.assign(category=annotation_config[label]) for a in scene.annotations],  # noqa: E501
                annotation_config=annotation_config
            ) for scene, label in zip(collection.scenes, labels)
        ],
        annotation_config=annotation_config
    )


def load_icdar2015(
    subset: str='train',
    text_category: str='text'
) -> SceneCollection:
    """Loads dataset from 2015 Robust Reading Competition.
    More details available at http://rrc.cvc.uab.es/?ch=4&com=introduction

    Args:
        subset: One of `train`, `test`, or `traintest`.
            If `traintest`, the scene collection contains both the
            train and test sets. If `train`, only the
            training set. If `test` only the test set.
        text_category: The category name to use for the
            annotation configuration.

    Returns:
        A scene collection containing the ICDAR 2015 dataset
        dataset.
    """

    scenes = []
    annotation_config = AnnotationConfiguration([text_category])

    def annotation_from_box(box):
        # We have to do this because there are no quotes
        # around text that contains commas in the annotation files.
        # For an example, see gt_img_115.txt.
        elements = box.split(',')
        x1, y1, x2, y2, x3, y3, x4, y4 = map(int, elements[:8])
        text = ','.join(elements[8:])  # noqa: F841
        return Annotation(
            selection=Selection(np.array([
                [x1, y1],
                [x2, y2],
                [x3, y3],
                [x4, y4]
            ])),
            category=annotation_config[text_category]
        )

    def scene_from_files(annotation_file, image_file):
        with open(annotation_file, 'r', encoding='utf-8-sig') as f:
            annotations = [
                annotation_from_box(box) for box in f.read().split('\n')[:-1]
            ]
        return Scene(
            image=image_file,
            annotations=annotations,
            annotation_config=annotation_config
        )

    def scenes_from_directories(annotations_dir, images_dir):
        pattern = re.compile(r'img_([0-9]+).')
        image_files = sorted(
            glob(path.join(images_dir, '*.jpg')),
            key=lambda x: pattern.findall(x)[0]
        )
        annotation_files = sorted(
            glob(path.join(annotations_dir, '*.txt')),
            key=lambda x: pattern.findall(x)[0]
        )
        assert len(image_files) == len(annotation_files), \
            'An error occurred loading the dataset.'
        p = utils.Progbar(
            len(annotation_files),
            task_name='Creating scenes from ' + annotations_dir
        )
        scenes = []
        for i, (annotation_file, image_file) in enumerate(
            zip(annotation_files, image_files)
        ):
            p.update(i + 1)
            scenes.append(
                scene_from_files(annotation_file, image_file)
            )
        return scenes

    if subset in ['train', 'traintest']:
        images_dir = utils.get_file(
            origin='https://storage.googleapis.com/miradata/datasets/rrc2015/ch4_training_images.zip',  # noqa: E501
            file_hash='07e2d816dab67df4cd509641b1e1e8c720d3b15873a65b6dc9c1aa6a4de3d2c6',  # noqa: E501
            cache_subdir=path.join('datasets', 'rrc2015'),
            hash_algorithm='sha256',
            extract=True,
            archive_format='zip',
            extract_check_fn=lambda directory: len(glob(path.join(directory, '*.jpg'))) == 1000  # noqa: E501
        )
        annotations_dir = utils.get_file(
            origin='https://storage.googleapis.com/miradata/datasets/rrc2015/ch4_training_localization_transcription_gt.zip',  # noqa: E501
            file_hash='b9f2f0343d016a326bcafe3c28e0ddbda18b3a1e4a8a595784ba9dc3e305d754',  # noqa: E501
            cache_subdir=path.join('datasets', 'rrc2015'),
            hash_algorithm='sha256',
            extract=True,
            archive_format='zip',
            extract_check_fn=lambda directory: len(glob(path.join(directory, '*.txt'))) == 1000  # noqa: E501
        )
        scenes += scenes_from_directories(
            annotations_dir,
            images_dir
        )
    if subset in ['test', 'traintest']:
        images_dir = utils.get_file(
            origin='https://storage.googleapis.com/miradata/datasets/rrc2015/ch4_test_images.zip',  # noqa: E501
            file_hash='ecfb2488333372b7381b14ccb5ac2127de88fb935f6759fe89e5c041b0e87358',  # noqa: E501
            cache_subdir=path.join('datasets', 'rrc2015'),
            hash_algorithm='sha256',
            extract=True,
            archive_format='zip',
            extract_check_fn=lambda directory: len(glob(path.join(directory, '*.jpg'))) == 500  # noqa: E501
        )
        annotations_dir = utils.get_file(
            origin='https://storage.googleapis.com/miradata/datasets/rrc2015/Challenge4_Test_Task1_GT.zip',  # noqa: E501
            file_hash='8f434c410cfcf680b8421e3c22abf31b9297d8e255c3b13acc2890fddc6db6b0',  # noqa: E501
            cache_subdir=path.join('datasets', 'rrc2015'),
            hash_algorithm='sha256',
            extract=True,
            archive_format='zip',
            extract_check_fn=lambda directory: len(glob(path.join(directory, '*.txt'))) == 500  # noqa: E501
        )
        scenes += scenes_from_directories(
            annotations_dir,
            images_dir
        )
    return SceneCollection(
        annotation_config=annotation_config,
        scenes=scenes
    )



from typing import Tuple, List
import logging

from tensorflow.keras import layers, regularizers, models, optimizers
from tensorflow.keras import backend as K
from nms import nms
import numpy as np
import cv2

from .. import core
from .detector import Detector

log = logging.getLogger(__name__)


class ScaleShift(layers.Layer):
    """Implement Scale / Shift opertaion from PVANet paper.

        out = in * gamma + beta,

    'gamma' and 'beta' are the learned weights and biases.

    Args:
        weights: Initialization weights.
            List of 2 Numpy arrays, with shapes:
            `[(input_shape,), (input_shape,)]`
    """
    def __init__(self, weights=None, **kwargs):
        self.initial_weights = weights
        super(ScaleShift, self).__init__(**kwargs)

    def build(self, input_shape):
        shape = (int(input_shape[-1]),)

        # Tensorflow >= 1.0.0 compatibility
        self._gamma = K.variable(np.ones(shape), name='{}_gamma'.format(self.name))  # noqa: E501
        self._beta = K.variable(np.zeros(shape), name='{}_beta'.format(self.name))  # noqa: E501
        self._trainable_weights = [self._gamma, self._beta]

        if self.initial_weights is not None:
            self.set_weights(self.initial_weights)
            del self.initial_weights

    def call(self, x, mask=None):
        return self._gamma * x + self._beta


def inception(
    x,
    f1x1: int,
    f3x3: Tuple[int, int],
    f5x5: Tuple[int, int, int],
    fOut: int,
    prefix: str,
    fPool: int=None,
    stride: int=1,
    outputBn=False
):
    """Make Inception layer according to the description in Figure 1
    in https://arxiv.org/pdf/1608.08021.pdf.

    """
    assert stride == 1 or fPool is not None, \
        'If stride != 1, fPool must be provided'
    assert (
        fOut == x.get_shape().as_list()[-1] or
        stride != 1
    ), 'The number of input filters must equal output filters when stride is 1.'  # noqa: E501
    x = layers.BatchNormalization(name=prefix+'/incep/bn')(x)
    x = layers.Activation('relu', name=prefix+'/incep/relu')(x)

    x1x1 = layers.Conv2D(f1x1, 1, strides=stride, name=prefix+'/incep/0/conv', use_bias=False)(x)  # noqa: E501
    x1x1 = layers.BatchNormalization(name=prefix+'/incep/0/bn')(x1x1)
    x1x1 = layers.Activation('relu', name=prefix+'/incep/0/relu')(x1x1)

    x3x3 = layers.Conv2D(f3x3[0], 1, strides=stride, name=prefix+'/incep/1_reduce/conv', use_bias=False)(x)  # noqa: E501
    x3x3 = layers.BatchNormalization(name=prefix+'/incep/1_reduce/bn')(x3x3)
    x3x3 = layers.Activation('relu', name=prefix+'/incep/1_reduce/relu')(x3x3)
    x3x3 = layers.Conv2D(f3x3[0], 3, name=prefix+'/incep/1_0/conv', padding='same', use_bias=False)(x3x3)  # noqa: E501
    x3x3 = layers.BatchNormalization(name=prefix+'/incep/1_0/bn')(x3x3)
    x3x3 = layers.Activation('relu', name=prefix+'/incep/1_0/relu')(x3x3)

    x5x5 = layers.Conv2D(f5x5[0], 1, strides=stride, name=prefix+'/incep/2_reduce/conv', use_bias=False)(x)  # noqa: E501
    x5x5 = layers.BatchNormalization(name=prefix+'/incep/2_reduce/bn')(x5x5)
    x5x5 = layers.Activation('relu', name=prefix+'/incep/2_reduce/relu')(x5x5)
    x5x5 = layers.Conv2D(f5x5[1], 3, name=prefix+'/incep/2_0/conv', padding='same', use_bias=False)(x5x5)  # noqa: E501
    x5x5 = layers.BatchNormalization(name=prefix+'/incep/2_0/bn')(x5x5)
    x5x5 = layers.Activation('relu', name=prefix+'/incep/2_0/relu')(x5x5)
    x5x5 = layers.Conv2D(f5x5[2], 3, name=prefix+'/incep/2_1/conv', padding='same', use_bias=False)(x5x5)  # noqa: E501
    x5x5 = layers.BatchNormalization(name=prefix+'/incep/2_1/bn')(x5x5)
    x5x5 = layers.Activation('relu', name=prefix+'/incep/2_1/relu')(x5x5)

    if stride == 2:
        xPool = layers.MaxPooling2D(3, strides=2, name=prefix+'/incep/pool', padding='same')(x)  # noqa: E501
        xPool = layers.Conv2D(fPool, 1, name=prefix+'/incep/poolproj/conv', use_bias=False)(xPool)  # noqa: E501
        xPool = layers.BatchNormalization(name=prefix+'/incep/poolproj/bn')(xPool)  # noqa: E501
        xPool = layers.Activation('relu', name=prefix+'/incep/poolproj/relu')(xPool)  # noqa: E501
        outs = [x1x1, x3x3, x5x5, xPool]
    elif stride == 1:
        outs = [x1x1, x3x3, x5x5]
    else:
        raise NotImplementedError

    y = layers.Concatenate(name=prefix+'/incep/concat')(outs)
    y = layers.Conv2D(fOut, 1, name=prefix+'/out/conv')(y)

    xRes = x
    if stride != 1:
        xRes = layers.Conv2D(
            fOut, 1, strides=stride, name=prefix+'/proj'
        )(xRes)
    y = layers.Add(name=prefix+'/add')([y, xRes])
    if outputBn:
        y = layers.BatchNormalization(name=prefix+'/last_bn')(y)
        y = layers.Activation('relu', name=prefix+'/last_relu')(y)
    return y


def crelu(
    x, k: int, fcReLU: int, prefix: str, stride: int=1,
    fIn: int=None, fOut: int=None, residual: bool=True,
    input_bn=True
):
    """Make c.ReLU according to the description in Figure 1 in
    https://arxiv.org/pdf/1608.08021.pdf.

    Args:
        x: The input layer
        k: The kernel size for the c.ReLU block
        fcReLU: The number of filters for the c.ReLU block
        prefix: The prefix for the layers in the block
        stride: The stride for the c.ReLU block
        fIn: The number of filters on the initial 1x1 convolution
        fOut: The number of filters for the output 1x1 convolution
        residual: If True, the input layer (x) is added as a residual.
        input_bn: Whether to apply batch normalization and activation
            to input.

    Returns:
        The output layer
    """
    xRes = x
    if input_bn:
        x = layers.BatchNormalization(name=prefix+'/1/bn')(x)
        x = layers.Activation('relu', name=prefix+'/1/relu')(x)
    if fIn is not None:
        x = layers.Conv2D(fIn, 1, name=prefix+'/1/conv')(x)
        x = layers.BatchNormalization(name=prefix+'/2/bn')(x)
        x = layers.Activation('relu', name=prefix+'/2/relu')(x)
    convolution = layers.Conv2D(
        fcReLU, (k, k), strides=stride,
        padding='same', use_bias=False, name=prefix+'/2/conv'
    )(x)
    batchnorm = layers.BatchNormalization(
        name=prefix+'/3/bn', center=False, scale=False
    )(convolution)
    negation = layers.Lambda(lambda x: -1*x, name=prefix+'/3/neg')(batchnorm)
    concatenation = layers.Concatenate(name=prefix+'/3/concat')(
        [batchnorm, negation]
    )
    scaleshift = ScaleShift(name=prefix+'3/scale')(concatenation)
    x = layers.Activation('relu', name=prefix+'/3/relu')(scaleshift)
    if fOut is not None:
        x = layers.Conv2D(fOut, 1, name=prefix+'/3/conv')(x)
        if residual:
            if stride != 1 or fOut != xRes.get_shape().as_list()[-1]:
                xRes = layers.Conv2D(fOut, 1, strides=stride, name=prefix+'/proj')(xRes)  # noqa: E501
            x = layers.Add(name=prefix+'/add')([x, xRes])
    return x


def build_backbone(input_shape=(1056, 640, 3)):
    """Build PVANet Backbone. The names of layers are intended to match
    those in
    https://github.com/sanghoon/pva-faster-rcnn/tree/master/models/pvanet
    """
    input_layer = layers.Input(input_shape)
    conv1_1 = crelu(input_layer, 7, 16, 'conv1_1', 2, residual=False)
    pool1_1 = layers.MaxPooling2D(3, 2, name='pool1_1', padding='same')(conv1_1)  # noqa: E501
    conv2_1 = crelu(pool1_1, 3, 24, 'conv2_1', fIn=24, fOut=64, input_bn=False)
    conv2_2 = crelu(conv2_1, 3, 24, 'conv2_2', fIn=24, fOut=64)
    conv2_3 = crelu(conv2_2, 3, 24, 'conv2_3', fIn=24, fOut=64)
    conv3_1 = crelu(conv2_3, 3, 48, 'conv3_1', 2, fIn=48, fOut=128)
    conv3_2 = crelu(conv3_1, 3, 48, 'conv3_2', fIn=48, fOut=128)
    conv3_3 = crelu(conv3_2, 3, 48, 'conv3_3', fIn=48, fOut=128)
    conv3_4 = crelu(conv3_3, 3, 48, 'conv3_4', fIn=48, fOut=128)
    conv4_1 = inception(conv3_4, 64, (48, 128), (24, 48, 48), 256, 'conv4_1', 128, 2)  # noqa: E501
    conv4_2 = inception(conv4_1, 64, (64, 128), (24, 48, 48), 256, 'conv4_2')
    conv4_3 = inception(conv4_2, 64, (64, 128), (24, 48, 48), 256, 'conv4_3')
    conv4_4 = inception(conv4_3, 64, (64, 128), (24, 48, 48), 256, 'conv4_4')
    conv5_1 = inception(conv4_4, 64, (96, 192), (32, 64, 64), 384, 'conv5_1', 128, 2)  # noqa: E501
    conv5_2 = inception(conv5_1, 64, (96, 192), (32, 64, 64), 384, 'conv5_2')
    conv5_3 = inception(conv5_2, 64, (96, 192), (32, 64, 64), 384, 'conv5_3')
    conv5_4 = inception(conv5_3, 64, (96, 192), (32, 64, 64), 384, 'conv5_4', outputBn=True)  # noqa: E501
    downscale = layers.MaxPooling2D(3, strides=2, name='downscale', padding='same')(conv3_4)  # noqa: E501
    upscale = layers.UpSampling2D(size=2, data_format='channels_last', interpolation='bilinear', name='upscale')(conv5_4)  # noqa: E501
    concat = layers.Concatenate(name='concat')([downscale, conv4_4, upscale])
    convf = layers.Conv2D(512, 1, name='convf')(concat)
    model = models.Model(inputs=input_layer, outputs=convf)
    return model


def build_model(backbone):
    """Build EAST Text Detection model on top of ResNet backbone.

    Args:
        backbone: The PVANet model built using build_backbone

    Returns:
        Two models, described below.

        The first is the EAST model with 2 outputs. All outputs
        bounded between 0 and 1 (sigmoid activation).

        Output #1 has shape (B, H // 4, W // 4, 1). It is
            is objectness score for each 4x4 box in the input image.
        Output #2 has shape (B, H // 4, W // 4, 5). It is
            the y1 (top), x2 (right), y2 (bottom), x1 (left), and angle
            offset for each predicted box from the input.

        The second model is the backbone for the EAST model (i.e.
        ResNet 50).
    """
    input_layer = backbone.input
    endpoint_names = reversed([
        'conv3_1/1/relu',
        'conv4_1/incep/relu',
        'conv5_1/incep/relu',
        'conv5_4/last_relu'
    ])
    endpoints = [backbone.get_layer(name) for name in endpoint_names]
    num_outputs = [None, 128, 64, 32]
    x = None
    for i, (filters, endpoint) in enumerate(zip(num_outputs, endpoints)):  # noqa: E501
        if filters is not None:
            x = layers.Concatenate()([x, endpoint.output])
            x = layers.Conv2D(filters, (1, 1), padding='same', kernel_regularizer=regularizers.l2(1e-5))(x)  # noqa: E501
            x = layers.BatchNormalization(momentum=0.997, epsilon=1e-5, scale=True)(x)  # noqa: E501
            x = layers.Activation('relu')(x)
            x = layers.Conv2D(filters, (3, 3), padding='same', kernel_regularizer=regularizers.l2(1e-5))(x)  # noqa: E501
            x = layers.BatchNormalization(momentum=0.997, epsilon=1e-5, scale=True)(x)  # noqa: E501
            x = layers.Activation('relu')(x)
        else:
            x = endpoint.output
        if i < 3:
            x = layers.UpSampling2D(
                size=(2, 2),
                interpolation='bilinear',
                name='resize_{0}'.format(i+1)
            )(x)
    x = layers.Conv2D(32, (3, 3), padding='same', kernel_regularizer=regularizers.l2(1e-5))(x)  # noqa: E501
    x = layers.BatchNormalization(momentum=0.997, epsilon=1e-5, scale=True)(x)
    x = layers.Activation('relu')(x)

    pred_score_map = layers.Conv2D(1, (1, 1), activation='sigmoid', name='pred_score_map')(x)  # noqa: E501
    rbox_geo_map = layers.Conv2D(4, (1, 1), activation='sigmoid', name='rbox_geo_map')(x)  # noqa: E501
    angle_map = layers.Conv2D(1, (1, 1), activation='sigmoid', name='rbox_angle_map')(x)  # noqa: E501
    angle_map = layers.Lambda(lambda x: (x - 0.5) * np.pi / 2)(angle_map)
    pred_geo_map = layers.Concatenate(axis=3, name='pred_geo_map')([rbox_geo_map, angle_map])  # noqa: E501
    outputs = layers.Concatenate(axis=3)([pred_score_map, pred_geo_map])
    return models.Model(inputs=[input_layer], outputs=outputs)


def loss(y_true, y_pred):
    # Compute score loss
    st, sp = [y[:, :, :, 0] for y in [y_true, y_pred]]  # noqa: E501
    beta = 1 - K.mean(st)
    score_loss = -(beta*st*K.log(sp + K.epsilon()) + (1-beta)*(1-st)*K.log(1. - sp + K.epsilon()))  # noqa: E501

    # Compute geo loss
    d1p, d2p, d3p, d4p = [y_pred[:, :, :, i:i+1] for i in range(1, 5)]
    d1t, d2t, d3t, d4t = [y_true[:, :, :, i:i+1] for i in range(1, 5)]
    thetap, thetat = [y[:, :, :, 5] for y in [y_pred, y_true]]
    wi = K.min(K.concatenate([d2p, d2t], axis=3), axis=3) + K.min(K.concatenate([d4p, d4t], axis=3), axis=3)  # noqa: E501
    hi = K.min(K.concatenate([d1p, d1t], axis=3), axis=3) + K.min(K.concatenate([d3p, d3t], axis=3), axis=3)  # noqa: E501
    intersection = wi*hi
    union = ((d2p + d4p)*(d1p + d3p) + (d2t + d4t)*(d1t + d3t))[:, :, :, 0] - intersection + K.epsilon()  # noqa: E501
    laabb = K.switch(K.equal(st, 1), -K.log((intersection / union) + K.epsilon()), K.zeros_like(intersection))  # noqa: E501
    ltheta = K.switch(K.equal(st, 1), K.abs(1 - K.cos(thetap - thetat)), K.zeros_like(thetap))  # noqa: E501
    geo_loss = laabb + 10*ltheta
    return score_loss + 1*geo_loss


class EAST(Detector):
    """A detector wrapping the
    `EAST Text Detector <https://arxiv.org/abs/1704.03155>`_. This version
    uses an implementation of PVANET for feature extraction,
    as in the original paper.

    Args:
        annotation_config: The annotation configuration to use for detection.
            It must contain `text` as one of the categories.
        input_shape: Tuple of (height, width, n_channels)
        pretrained_backbone: Whether to use a pretrained backbone for the model
        pretrained_top: Whether to use the pretrained full model.
        text_category: The name of the category to use for labeling text.

    Attributes:
        model: The underlying model
        backbone: The backbone for the model (i.e., ResNet50)
    """
    def __init__(
        self,
        annotation_config: core.AnnotationConfiguration=None,
        input_shape: Tuple[int, int]=(None, None, 3),
        pretrained_backbone: bool=False,
        pretrained_top: bool=False,
        text_category='text'
    ):
        if annotation_config is None:
            annotation_config = core.AnnotationConfiguration(['text'])
        assert text_category in annotation_config, \
            'This model only detects text, so, at a minimum, the provided `text_category` must be in annotation_config'  # noqa; E501
        assert len(input_shape) == 3, \
            'Input shape must have shape (h, w, n_channels).'
        if pretrained_top and pretrained_backbone:
            log.warning(
                'pretrained_top makes pretrained_backbone '
                'redundant. Disabling pretrained_backbone.'
            )
            pretrained_backbone = False
        self.annotation_config = annotation_config
        self.text_category = self.annotation_config[text_category]
        self.backbone = build_backbone(input_shape)

        self.model = build_model(self.backbone)
        if pretrained_top or pretrained_backbone:
            raise NotImplementedError('Pretrained weights not yet available.')
        self.compile()

    def invert_targets(
        self,
        y,
        images,
        threshold=0.8,
        **nms_kwargs
    ):
        scenes = []
        for yi, image in zip(y, images):
            scores = yi[:, :, 0:1]
            geo = yi[:, :, 1:]
            # (N, 2): List of coordinates above threshold in
            # the output resolution. The values are (xOut, yOut)
            # and the list is sorted in order of ascending yOut.
            xy_out = np.argwhere(scores[:, :, 0] > threshold)[:, ::-1]

            N = len(xy_out)

            # (N, 2): List of coordinates above threshold
            # in the input resolution. We need only multiply
            # xy_out by 4 to recover the original because
            # the model output map resolution is 1 / 4 of
            # input resolution). The values are (xIn, yIn)
            # and the list is sorted in order of ascending yIn.
            origin = xy_out*4

            # (N, 5): List of adjustments to the raw boxes
            # where the values are (d_top, d_right,
            # d_bottom, d_left, theta)
            geometry = geo[xy_out[:, 1], xy_out[:, 0], :]
            geometry[:, 0] *= image.shape[0]
            geometry[:, 1] *= image.shape[1]
            geometry[:, 2] *= image.shape[0]
            geometry[:, 3] *= image.shape[1]

            boxes = np.zeros((N, 9))
            boxes[:, 8] = scores[scores[:, :, 0] > threshold, 0]
            if len(boxes) == 0:
                annotations = []
            else:
                for i, (xc, yc), (dt, dr, db, dl, theta) in zip(
                    range(N),
                    origin,
                    geometry
                ):
                    theta *= 180/np.pi
                    # Assume x1, y1
                    X = np.array([
                        [0, -dt - db],        # p1 (bottom right)
                        [dr + dl, -dt - db],  # p2 (top right)
                        [dr + dl, 0],         # p3 (top left)
                        [0, 0],               # p4 (bottom left)
                        [dl, -db]             # p5 (anchor offset)
                    ])
                    # According to paper, rotation is about the
                    # bottom left corner, corresponding to x1, y2.
                    M = cv2.getRotationMatrix2D(
                        center=(0, 0),
                        angle=theta,
                        scale=1
                    )
                    X = cv2.transform(np.array([X]), M)[0]
                    X[4, 0] = xc - X[4, 0]
                    X[4, 1] = yc - X[4, 1]
                    X[:4, :] += X[4]
                    boxes[i, :8] = X[:4].reshape(-1)
                nmsIdx = nms.polygons(
                    polys=boxes[:, :8].reshape(-1, 4, 2),
                    scores=boxes[:, 8],
                    **nms_kwargs
                )
                boxes = boxes[nmsIdx]
                annotations = [
                    core.Annotation(
                        category=self.text_category,
                        selection=core.Selection(points=box[:8].reshape((4, 2))),  # noqa: E501
                        score=box[-1]
                    ) for box in boxes
                ]
            scenes.append(
                core.Scene(
                    annotation_config=self.annotation_config,
                    annotations=annotations,
                    image=image
                )
            )
        return scenes

    def compute_targets(
        self,
        collection: core.SceneCollection
    ):
        # y has shape (B, H // 4, W // 4, 6)
        # The entries are score, dt, dr, db, dl, theta
        ys = []
        for scene in collection:
            image_h = scene.image.height
            image_w = scene.image.width
            cy = np.arange(start=0, stop=image_h, dtype='float32', step=4)
            cx = np.arange(start=0, stop=image_w, dtype='float32', step=4)
            xy = np.concatenate([
                np.tile(cx, len(cy))[:, np.newaxis],
                np.repeat(cy, len(cx))[:, np.newaxis]
            ], axis=1)
            scores = np.zeros((len(xy), 1))
            geometry = np.zeros((len(xy), 5))

            for ann in scene.annotations:
                if ann.category != self.text_category:
                    continue
                ((xtl, ytl), (xtr, ytr), (xbr, ybr), (xbl, ybl)), theta = ann.selection.rbox()  # noqa; E501
                positive = ann.selection.contains_points(xy)
                origin = xy[positive]

                # Horizontal distance
                lh = np.sqrt((ybr - ybl) ** 2 + (xbr - xbl) ** 2)

                # Vertical distance
                lv = np.sqrt((ytl - ybl) ** 2 + (xtl - xbl) ** 2)

                dt = abs(
                    (ytr - ytl)*origin[:, 0] -
                    (xtr - xtl)*origin[:, 1] +
                    xtr*ytl - ytr*xtl
                ) / lh
                db = lv - dt

                dr = abs(
                    (ytr - ybr)*origin[:, 0] -
                    (xtr - xbr)*origin[:, 1] +
                    xtr*ybr - ytr*xbr
                ) / lv
                dl = lh - dr
                scores[positive] = 1

                geometry[positive, 0] = dt / image_h  # dt
                geometry[positive, 1] = dr / image_w  # dr
                geometry[positive, 2] = db / image_h  # db
                geometry[positive, 3] = dl / image_w  # dl
                geometry[positive, 4] = theta

            dt = geometry[:, 0]
            dl = geometry[:, 1]
            lv = geometry[:, 0] + geometry[:, 2]
            lh = geometry[:, 1] + geometry[:, 3]
            dlr = np.divide(dl, lh, out=np.zeros_like(dl), where=lh != 0)
            dtr = np.divide(dt, lv, out=np.zeros_like(dt), where=lv != 0)

            # Only include shrunken boxes
            margin = 0.15
            exclude = (
                (dlr <= margin) | (dlr >= 1-margin) |
                (dtr <= margin) | (dtr >= 1-margin)
            ) & (scores[:, 0] == 1)
            scores[exclude] = 0

            y1i = scores.reshape((len(cy), len(cx), 1))
            y2i = geometry.reshape((len(cy), len(cx), 5))
            ys.append(np.concatenate([y1i, y2i], axis=2))
        return np.float32(ys)

    def compute_inputs(self, images: List[core.Image]):
        assert all(
            [all(s % 32 == 0 for s in image.shape[:2]) for image in images]
        ), \
            'Input shapes must all be multiples of 32.'
        assert all(
            image.shape[0] == image.shape[1] for image in images
        ), 'All images must be square.'
        return np.float32([
            image.scaled(-1, 1)
            for image in images
        ])

    def compile(self):
        self.model.compile(
            loss=loss,
            optimizer=optimizers.Adam()
        )



from typing import Tuple, List
from os import path
import logging

from keras_retinanet import losses as rn_losses
from keras_retinanet.models import backbone as rn_backbone
from keras_retinanet.utils import anchors as rn_anchors
from keras import layers, optimizers
import numpy as np
import cv2

from .detector import Detector
from .. import core
from .. import utils

log = logging.getLogger(__name__)

pretrained = {
    'resnet50': {
        'hash':
        '6518ad56a0cca4d1bd8cbba268dd4e299c7633efe7d15902d5acbb0ba180027c',  # noqa: E501
        'url':
        'https://storage.googleapis.com/miradata/weights/retinanet/resnet50_coco_best_v2.1.0.h5'  # noqa: E501
    }
}


class RetinaNet(Detector):
    """A detector wrapping RetinaNet. All the heavy lifting is done by the
    `keras-retinanet <https://github.com/fizyr/keras-retinanet>`_ package.

    Args:
        annotation_config: The annotation configuration to use for detection
        input_shape: Tuple of (height, width, n_channels)
        pretrained_backbone: Whether to use a pretrained backbone for the model
        pretrained_top: Whether to use the pretrained full model (only
            supported for backbone `resnet50`)
        backbone_name: The name of the backbone to use. See `keras-retinanet`
            for valid options.

    Attributes:
        model: The base Keras model containing the weights for feature
            extraction and bounding box regression / classification model.
            You should use this model for loading and saving weights.
    """

    def __init__(
            self,
            annotation_config: core.AnnotationConfiguration = core.
            AnnotationConfiguration.COCO,  # noqa: E501
            input_shape: Tuple[int, int, int] = (None, None, 3),
            pretrained_backbone: bool = True,
            pretrained_top: bool = False,
            backbone_name='resnet50'):
        if pretrained_top and pretrained_backbone:
            log.info(
                'Disabling imagenet weights, using pretrained full network '
                'instead.')  # noqa: E501
            pretrained_backbone = False
        if pretrained_top:
            assert backbone_name in pretrained, (
                'The selected backbone has no fully trained network. '
                'Acceptable options are: {0}. '  # noqa: E501
                'To resolve this error, choose set pretrained_top=False.'
            ).format(','.join(list(pretrained.keys())))
        self.anchor_params = rn_anchors.AnchorParameters.default
        self.annotation_config = annotation_config
        self.rn_backbone = rn_backbone(backbone_name)
        self.model = self.rn_backbone.retinanet(
            inputs=layers.Input(input_shape),
            num_classes=len(annotation_config),
            num_anchors=self.anchor_params.num_anchors())
        if pretrained_backbone:
            weights_path = self.rn_backbone.download_imagenet()
            log.info('Loading weights from ' + weights_path)
            self.model.load_weights(weights_path, by_name=True)
        if pretrained_top:
            assert annotation_config == core.AnnotationConfiguration.COCO, \
                'To use pretrained_top, annotation config must be core.AnnotationConfiguration.COCO'  # noqa: E501
            pretrained_params = pretrained[backbone_name]
            weights_path = utils.get_file(origin=pretrained_params['url'],
                                          file_hash=pretrained_params['hash'],
                                          cache_subdir=path.join(
                                              'weights', 'retinanet'),
                                          hash_algorithm='sha256',
                                          extract=False)
            self.model.load_weights(weights_path, by_name=True)
        self.compile()

    def invert_targets(self, y, images, threshold=0.5, nms_threshold=0.1):

        boxes = rn_anchors.anchors_for_shape(image_shape=images[0].shape,
                                             anchor_params=self.anchor_params)

        # We filter the third dimension in order to remove the anchor states
        # if `invert_targets` is directly called with the output
        # of `compute_targets`.
        y0 = y[0][:, :, :4]
        y1 = y[1][:, :, :len(self.annotation_config)]

        # The code below reproduces the logic from bbox_transform_inv
        # https://github.com/fizyr/keras-retinanet/blob/master/keras_retinanet/backend/common.py
        # The only changes are to:
        # * Remove the use of keras because we are not computing the
        #   boxes as part of the network.
        # * Remove the batch dimension of the boxes because they are
        #   from the `anchors_for_shape` function and not part of the
        #   network.

        mean = [0, 0, 0, 0]
        std = [0.2, 0.2, 0.2, 0.2]

        width = boxes[:, 2] - boxes[:, 0]
        height = boxes[:, 3] - boxes[:, 1]

        xl = boxes[:, 0] + (y0[:, :, 0] * std[0] + mean[0]) * width
        yt = boxes[:, 1] + (y0[:, :, 1] * std[1] + mean[1]) * height
        xr = boxes[:, 2] + (y0[:, :, 2] * std[2] + mean[2]) * width
        yb = boxes[:, 3] + (y0[:, :, 3] * std[3] + mean[3]) * height
        w = xr - xl
        h = yb - yt
        xywh = np.concatenate([np.expand_dims(x, -1) for x in [xl, yt, w, h]],
                              axis=2)

        scenes = []
        for boxes, labels, image in zip(xywh, y1, images):
            positive = labels.max(axis=1) > threshold
            if positive.sum() == 0:
                annotations = []
            else:
                boxes = boxes[positive]
                labels = labels[positive]
                classes = np.where(labels.max(axis=0) > threshold)[0]
                predictions = []
                for c in set(classes):
                    class_positive = labels[:, c] > threshold
                    subboxes = boxes[class_positive]
                    sublabels = labels[class_positive, c]
                    bestIdxs = cv2.dnn.NMSBoxes(bboxes=subboxes.tolist(),
                                                scores=sublabels.tolist(),
                                                score_threshold=threshold,
                                                nms_threshold=0.1)[:, 0]
                    subboxes = subboxes[bestIdxs]
                    predictions.extend([b.tolist() + [c] for b in subboxes])
                    annotations = [
                        core.Annotation(selection=core.Selection(
                            [[x, y], [x + w, y + h]]),
                                        category=self.annotation_config[c])
                        for x, y, w, h, c in predictions
                    ]
            scenes.append(
                core.Scene(annotations=annotations,
                           annotation_config=self.annotation_config,
                           image=image))
        return scenes

    def compute_targets(self, collection: core.SceneCollection):
        # Targets are two outputs with shapes
        # (B, N, 5) (four anchor regression targets + anchor state)
        # (B, N, M + 1) (M classification targets + anchor state)
        assert collection.annotation_config == self.annotation_config, \
            'Found incompatible annotation configuration.'
        annotations_group = [{
            'bboxes': b[:, :4],
            'labels': b[:, 4]
        } for b in [s.bboxes() for s in collection]]
        images = collection.images
        anchors = rn_anchors.anchors_for_shape(
            image_shape=images[0].shape, anchor_params=self.anchor_params)
        return list(
            rn_anchors.anchor_targets_bbox(anchors=anchors,
                                           image_group=images,
                                           annotations_group=annotations_group,
                                           num_classes=len(
                                               self.annotation_config),
                                           negative_overlap=0.4,
                                           positive_overlap=0.5))

    def compute_inputs(self, images: List[core.Image]):
        return np.float32(
            [self.rn_backbone.preprocess_image(image) for image in images])

    def compile(self):
        self.model.compile(loss={
            'regression': rn_losses.smooth_l1(),
            'classification': rn_losses.focal()
        },
                           optimizer=optimizers.adam(lr=1e-5, clipnorm=0.001))

    def freeze_backbone(self):
        output_names = [
            'regression_submodel', 'classification_submodel', 'P3', 'P4', 'P5',
            'P6', 'P7'
        ]
        for l in self.model.layers:
            if (l.name in output_names or l.name in self.model.output_names):
                log.info('Not freezing ' + l.name)
            else:
                log.info('Freezing ' + l.name)
                l.trainable = False
        self.compile()

    def unfreeze_backbone(self):
        """Unfreeze the body of the model, making all layers trainable."""
        for l in self.model.layers:
            l.trainable = True
        self.compile()



from .yolov3 import YOLOv3  # noqa: F401
from .retinanet import RetinaNet  # noqa: F401
from .east import EAST  # noqa: F401
from .advancedeast import AdvancedEAST  # noqa: F401
from .mtcnn import MTCNN  # noqa: F401
from .detector import Detector  # noqa: F401


from mira import detectors, core


annotation_config = core.AnnotationConfiguration(['foo', 'bar'])
collection = core.SceneCollection(
    annotation_config=annotation_config,
    scenes=[
        core.Scene(
            image=core.Image.new(
                width=100,
                height=100,
                channels=3
            ),
            annotation_config=annotation_config,
            annotations=[
                core.Annotation(
                    category=annotation_config['foo'],
                    selection=core.Selection([[0, 0], [50, 50]])
                )
            ]
        ),
        core.Scene(
            image=core.Image.new(
                width=100,
                height=100,
                channels=3
            ),
            annotation_config=annotation_config,
            annotations=[
                core.Annotation(
                    category=annotation_config['bar'],
                    selection=core.Selection([[20, 20], [80, 80]])
                )
            ]
        ),
    ]
)


def test_retinanet():
    rn = detectors.RetinaNet(
        pretrained_backbone=False,
        annotation_config=annotation_config
    )
    inverted = rn.invert_targets(y=rn.compute_targets(collection), images=collection.images)  # noqa: E501
    assert all(e - 1 <= a < e + 1 for a, e in zip(inverted[0].annotations[0].selection.bbox(), [0, 0, 50, 50]))  # noqa: E501
    assert all(e - 1 <= a < e + 1 for a, e in zip(inverted[1].annotations[0].selection.bbox(), [20, 20, 80, 80]))  # noqa: E501


def test_yolo():
    yolo = detectors.YOLOv3(
        pretrained_backbone=False,
        annotation_config=annotation_config
    )
    inverted = yolo.invert_targets(y=yolo.compute_targets(collection), images=collection.images)  # noqa: E501
    assert all(e - 1 <= a < e + 1 for a, e in zip(inverted[0].annotations[0].selection.bbox(), [0, 0, 50, 50]))  # noqa: E501
    assert all(e - 1 <= a < e + 1 for a, e in zip(inverted[1].annotations[0].selection.bbox(), [20, 20, 80, 80]))  # noqa: E501


def test_east():
    east = detectors.EAST(
        pretrained_backbone=False,
        annotation_config=annotation_config,
        text_category='foo'
    )
    inverted = east.invert_targets(y=east.compute_targets(collection), images=collection.images)  # noqa: E501
    assert all(e - 1 <= a < e + 1 for a, e in zip(inverted[0].annotations[0].selection.bbox(), [0, 0, 50, 50]))  # noqa: E501
    assert len(inverted[1].annotations) == 0


def test_advancedeast():
    advancedeast = detectors.AdvancedEAST(
        pretrained_backbone=False,
        annotation_config=annotation_config,
        text_category='foo'
    )
    inverted = advancedeast.invert_targets(y=advancedeast.compute_targets(collection), images=collection.images)  # noqa: E501
    assert all(e - 2 <= a < e + 2 for a, e in zip(inverted[0].annotations[0].selection.bbox(), [0, 0, 50, 50]))  # noqa: E501
    assert len(inverted[1].annotations) == 0




from os import path
from typing import List
from xml.etree import ElementTree
import logging

from ..utils import Progbar
from ..core import (
    Selection,
    Scene,
    SceneCollection,
    AnnotationConfiguration,
    Annotation
)

log = logging.getLogger(__name__)


VOC_SCENE_METADATA_MAP = [
    ['folder'],
    ['filename'],
    ['size', 'width'],
    ['size', 'height'],
    ['source', 'database'],
    ['source', 'annotation'],
    ['source', 'image'],
    ['segmented']
]

VOC_ANNOTATION_METADATA_MAP = [
    ['pose'],
    ['truncated'],
    ['difficult']
]


def map_xml_to_metadata(
    paths: List[List[str]],
    root: ElementTree.Element
):
    metadata = {}
    for p in paths:
        key = ':'.join(p)
        elem = root
        for part in p:
            elem = elem.find(part)
            if elem is None:
                log.info(
                    'Missing annotation metadata: {0}'.format(
                        ':'.join(p)
                    )
                )
                break
        if elem is None:
            continue
        metadata[key] = elem.text
    return metadata


def load_voc(
    filepaths: str,
    annotation_config: AnnotationConfiguration,
    image_dir: str=None,
) -> SceneCollection:
    """Read a scene from a VOC XML annotation file. Remaining arguments
    passed to scene constructor.

    Args:
        filepaths: A list of VOC files to read
        image_folder: Folder in which to look for images. Defaults to same
            folder as XML file prepended to the folder specified in the
            XML file.
        annotation_config: The annotation configuration to use.

    Returns:
        A new scene collection, one scene per VOC file
    """
    scenes = []
    p = Progbar(
        len(filepaths),
        task_name='Loading {0} VOC annotation files.'.format(len(filepaths))
    )
    for i, filepath in enumerate(filepaths):
        annotations = []
        root = ElementTree.parse(filepath).getroot()

        # Get the scene level metadata
        scene_metadata = map_xml_to_metadata(
            paths=VOC_SCENE_METADATA_MAP,
            root=root
        )

        if image_dir is None:
            folder = scene_metadata['folder']
            if folder is None:
                image_dir = path.dirname(filepath)
            else:
                image_dir = path.join(
                    path.dirname(filepath),
                    folder
                )

        image_path = path.join(image_dir, scene_metadata['filename'])
        for obj in root.findall('object'):
            category = annotation_config[obj.find('name').text]
            selection = None
            for bndbox in obj.findall('bndbox'):
                xmin, ymin, xmax, ymax = map(
                    lambda k: int(float(bndbox.find(k).text)),
                    ['xmin', 'ymin', 'xmax', 'ymax']
                )
                current = Selection(
                    [
                        [xmin, ymin],
                        [xmax, ymax]
                    ]
                )
                if selection is None:
                    selection = current
                else:
                    selection += current
            annotations.append(Annotation(
                selection=selection,
                category=category
            ))
        p.update(i+1)
        scenes.append(Scene(
            annotation_config=annotation_config,
            annotations=annotations,
            image=image_path
        ))
    return SceneCollection(
        scenes=scenes,
        annotation_config=annotation_config
    )



from typing import Tuple
from os import path
import logging
import itertools

from keras import applications, layers, models, optimizers
from keras import backend as K
import numpy as np
import cv2

from ..core import (
    Scene,
    Selection,
    Annotation,
    AnnotationConfiguration
)
from .detector import Detector
from .. import utils

log = logging.getLogger(__name__)


def build_backbone(input_shape=(None, None, 3), pretrained=False):
    input_layer = layers.Input(input_shape)
    vgg16 = applications.vgg16.VGG16(
        input_tensor=input_layer,
        weights='imagenet' if pretrained else None,
        include_top=False
    )
    features = [
        vgg16.get_layer('block{0}_pool'.format(i)).output for i in range(2, 6)
    ]
    return models.Model(inputs=input_layer, outputs=features)


def build_model(backbone):
    x = None
    for outputIdx in range(len(backbone.outputs)-1, 0, -1):
        hIdx = len(backbone.outputs) - outputIdx + 1
        filters = 2**(outputIdx + 4)
        if x is None:
            x = backbone.outputs[outputIdx]
        x = layers.UpSampling2D((2, 2), name='h{0}/upsample'.format(hIdx))(x)
        x = layers.Concatenate(name='h{0}/concat'.format(hIdx))([x, backbone.outputs[outputIdx - 1]])  # noqa: E501
        x = layers.BatchNormalization(name='h{0}/bn1'.format(hIdx))(x)
        x = layers.Conv2D(
            filters=filters, kernel_size=1, activation='relu', padding='same', name='h{0}/conv1'.format(hIdx)  # noqa: E501
        )(x)
        x = layers.BatchNormalization(name='h{0}/bn2'.format(hIdx))(x)
        x = layers.Conv2D(
            filters=filters, kernel_size=3, activation='relu', padding='same', name='h{0}/conv2'.format(hIdx)  # noqa: E501
        )(x)
    x = layers.BatchNormalization(name='features/bn')(x)
    x = layers.Conv2D(filters=32, kernel_size=3, activation='relu', padding='same', name='features/conv')(x)  # noqa: E501
    y1 = layers.Conv2D(filters=1, kernel_size=1, padding='same', name='inside_score', activation='sigmoid')(x)  # noqa: E501
    y2 = layers.Conv2D(filters=2, kernel_size=1, padding='same', name='side_vertex_code', activation='sigmoid')(x)  # noqa: E501
    y3 = layers.Conv2D(filters=4, kernel_size=1, padding='same', name='side_vertex_coord')(x)  # noqa: E501
    east_detect = layers.Concatenate(axis=-1, name='east_detect')([y1, y2, y3])
    return models.Model(inputs=backbone.inputs, outputs=east_detect, name='advanced_east')  # noqa: E501


def loss(y_true, y_pred):
    eps = K.epsilon()

    # Compute score loss
    st, sp = y_true[..., 0:1], y_pred[..., 0:1]
    beta = 1 - K.mean(st)
    score_loss = -(beta*st*K.log(sp + eps) + (1-beta)*(1-st)*K.log(1. - sp + eps))  # noqa: E501
    score_loss = K.mean(score_loss)

    # Compute end loss
    et, ep = y_true[..., 1:3], y_pred[..., 1:3]
    beta = 1 - (K.mean(et[..., 0]) / (K.mean(st) + eps))
    end_loss = -(beta*et*K.log(ep + eps) + (1-beta)*(1-et)*K.log(1. - ep + eps))  # noqa: E501
    end_loss = K.sum(end_loss * st) / (K.sum(st) + eps)

    # Compute vertex loss
    vt, vp = y_true[..., 3:], y_pred[..., 3:]
    sizes = 4*K.sqrt(K.sum(K.square(vt[..., 0:2] - vt[..., 2:4]), axis=-1)) + eps  # noqa: E501
    diff = K.abs(vp - vt)
    smooth_l1 = (K.sum(K.switch(K.less(diff, 1), 0.5 * K.square(diff), diff - 0.5), axis=-1) / sizes) * et[..., 0]  # noqa: E501
    geo_loss = K.sum(smooth_l1) / (K.sum(st) + eps)

    score_weight, end_weight, geo_weight = 4.0, 1, 1
    return score_weight*score_loss + end_weight*end_loss + geo_weight*geo_loss


class AdvancedEAST(Detector):
    """A detector wrapping the
    `Advanced EAST Text Detector <https://github.com/huoyijie/AdvancedEAST>`_.
    The original code was ported over. The pretrained weights also originate
    from the linked repository. This version of EAST is expected to perform
    better on longer text boxes relative to the input image.

    Args:
        annotation_config: The annotation configuration to use for detection.
            It must contain `text` as one of the categories.
        input_shape: Tuple of (height, width, n_channels)
        pretrained_backbone: Whether to use a pretrained backbone for the model
        pretrained_top: Whether to use the pretrained full model.
        text_category: The name of the category to use for labeling text.

    Attributes:
        model: The underlying model
        backbone: The backbone for the model (i.e., VGG16)
    """
    def __init__(
        self,
        annotation_config: AnnotationConfiguration=None,
        input_shape: Tuple[int, int]=(None, None, 3),
        pretrained_backbone: bool=True,
        pretrained_top: bool=False,
        text_category='text'
    ):
        if annotation_config is None:
            annotation_config = AnnotationConfiguration(['text'])
        assert text_category in annotation_config, \
            'This model only detects text, so, at a minimum, the provided `text_category` must be in annotation_config'  # noqa; E501
        assert len(input_shape) == 3, \
            'Input shape must have shape (h, w, n_channels).'
        if pretrained_top and pretrained_backbone:
            log.warning(
                'pretrained_top makes pretrained_backbone '
                'redundant. Disabling pretrained_backbone.'
            )
            pretrained_backbone = False
        self.annotation_config = annotation_config
        self.text_category = self.annotation_config[text_category]
        self.backbone = build_backbone(
            input_shape=input_shape,
            pretrained=pretrained_backbone
        )
        self.model = build_model(self.backbone)
        if pretrained_top:
            weights_path = utils.get_file(
                origin='https://storage.googleapis.com/miradata/weights/advancedeast/east_model_weights_3T736.h5',  # noqa; E501
                file_hash='1a5ef4e304cd71358fcfd4403859c38735738eed2961ea9b34484c7e4199d4f6',  # noqa; E501
                cache_subdir=path.join('weights', 'advancedeast'),
                hash_algorithm='sha256',
                extract=False
            )
            self.model.load_weights(weights_path)
        self.compile()

    def invert_targets(self, y, images, threshold=0.9):
        scenes = []
        placeholder = np.zeros(y.shape[1:-1])

        def contourCoords(contour):
            placeholder[...] = 0
            cv2.drawContours(
                image=placeholder,
                contours=[contour],
                contourIdx=-1,
                color=1,
                thickness=-1
            )
            yp, xp = np.where(placeholder > 0)
            return xp, yp

        def vertices_from_mask(mask, predictions):
            contours, _ = cv2.findContours(
                image=mask,
                mode=cv2.RETR_TREE,
                method=cv2.CHAIN_APPROX_SIMPLE
            )
            vertices = []
            for contour in contours:
                placeholder[...] = 0
                xc, yc = contourCoords(contour)
                w, t, dx1, dy1, dx2, dy2 = predictions[yc, xc, 1:].T
                w /= w.sum()
                x1, y1 = 4*xc + dx1, 4*yc + dy1
                x2, y2 = 4*xc + dx2, 4*yc + dy2
                x1, y1, x2, y2 = map(
                    lambda v: sum(w * v),
                    [x1, y1, x2, y2]
                )
                vertices.append([
                    [x1, y1],
                    [x2, y2]
                ])
            return vertices

        for yi, image in zip(y, images):
            mask = np.uint8(((yi[..., :3] > threshold) & (yi[..., 0:1] > threshold))*255)  # noqa: E501
            regionContours, _ = cv2.findContours(
                image=mask[..., 0],
                mode=cv2.RETR_TREE,
                method=cv2.CHAIN_APPROX_SIMPLE
            )
            boxes = []
            for regionContour in regionContours:
                xc, yc = contourCoords(regionContour)
                mVT = mask[..., 1:3]*placeholder[..., np.newaxis]
                mH = np.uint8(255*((mVT[..., 0] > 127.5) & (mVT[..., 1] < 127.5)))  # noqa: E501
                mT = np.uint8(255*((mVT[..., 0] > 127.5) & (mVT[..., 1] > 127.5)))  # noqa: E501
                if mH.max() == 0 or mT.max() == 0:
                    continue
                hVertices = vertices_from_mask(mH, yi)
                tVertices = vertices_from_mask(mT, yi)
                subcontours = [
                    (np.array(h + t).reshape(4, 1, 2) // 4).astype('int32') for h, t in itertools.product(hVertices, tVertices)  # noqa: E501
                ]
                areas = [
                    cv2.contourArea(c) for c in subcontours
                ]
                subcontour = subcontours[np.argmax(areas)]
                xs, ys = contourCoords(np.int32(subcontour))
                score = yi[ys, xs, 0].mean()
                boxes.append(
                    subcontour.reshape(-1).tolist() + [score]
                )
            boxes = np.float32(boxes)
            annotations = [
                Annotation(
                    selection=Selection(box[:8].reshape(4, 2)),
                    score=box[-1],
                    category=self.text_category
                ).resize(4) for box in boxes
            ]
            scenes.append(
                Scene(
                    image=image,
                    annotation_config=self.annotation_config,
                    annotations=annotations
                )
            )
        return scenes

    def compute_targets(self, collection):
        # y has shape (B, H // 4, W // 4, 7)
        # The entries are score, left, right, dx1, dy1, dx2, dy2
        B = len(collection)
        H = collection[0].image.height
        W = collection[0].image.width
        y = np.zeros((B, H // 4, W // 4, 7))
        points = np.mgrid[:y.shape[1], :y.shape[2]].reshape(2, -1).T
        for i, scene in enumerate(collection):
            for ann in scene.annotations:
                if ann.category != self.text_category:
                    continue
                selection = ann.selection.resize(0.25)
                xa, ya = points[selection.contains_points(points)].T
                ((xtl, ytl), (xtr, ytr), (xbr, ybr), (xbl, ybl)), theta = selection.rbox()  # noqa: E501

                # Horizontal distance
                lx = np.sqrt((ybr - ybl) ** 2 + (xbr - xbl) ** 2)

                # Vertical distance
                ly = np.sqrt((ytl - ybl) ** 2 + (xtl - xbl) ** 2)

                # Distance to top edge
                dy = abs(
                    (ytr - ytl)*xa -
                    (xtr - xtl)*ya +
                    xtr*ytl - ytr*xtl
                ) / lx

                # Distance to left edge
                dx = lx - abs(
                    (ytr - ybr)*xa -
                    (xtr - xbr)*ya +
                    xtr*ybr - ytr*xbr
                ) / ly

                # Scale distances between 0 and 1.
                # dx == 0 corresponds with the left edge
                # and dx == 1 corresponds with the right edge.
                dx /= lx
                dy /= ly

                # Positive scores are for the box with the radius
                # from 0 to 0.8. This corresponds to shrinking
                # the overall box by 0.1. So 0.1 -> 0.9 on the
                # overall box are positive.
                isPositive = (dx > 0.1) & (dx < 0.9) & (dy > 0.1) & (dy < 0.9)  # noqa: E501

                edgeMargin = 0.2

                # Head and tail are defined by longest edge.
                if ly > lx:
                    dLong = dy
                    xh1, yh1 = xtl, ytl
                    xh2, yh2 = xtr, ytr
                    xt1, yt1 = xbl, ybl
                    xt2, yt2 = xbr, ybr
                else:
                    dLong = dx
                    xh1, yh1 = xtl, ytl
                    xh2, yh2 = xbl, ybl
                    xt1, yt1 = xtr, ytr
                    xt2, yt2 = xbr, ybr

                isBorder = ((dLong < edgeMargin) | (dLong > 1 - edgeMargin)) & isPositive  # noqa: E501
                isTail = (dLong > (1 - edgeMargin)) & isPositive
                isHead = (dLong < edgeMargin) & isPositive

                y[i, ya[isPositive], xa[isPositive], 0] = 1
                y[i, ya[isBorder], xa[isBorder], 1] = 1
                y[i, ya[isTail], xa[isTail], 2] = 1
                y[i, ya[isHead], xa[isHead], 3:7] = np.concatenate([
                    v[:, np.newaxis] for v in [(xh1 - xa), (yh1 - ya), (xh2 - xa), (yh2 - ya)]  # noqa: E501
                ], axis=1)[isHead]*4
                y[i, ya[isTail], xa[isTail], 3:7] = np.concatenate([
                    v[:, np.newaxis] for v in [(xt1 - xa), (yt1 - ya), (xt2 - xa), (yt2 - ya)]  # noqa: E501
                ], axis=1)[isTail]*4
            return y

    def compute_inputs(self, images):
        assert all(
            [all(s % 32 == 0 for s in image.shape[:2]) for image in images]
        ), \
            'Input shapes must all be multiples of 32.'
        assert all(
            image.shape[0] == image.shape[1] for image in images
        ), 'All images must be square.'
        return applications.vgg16.preprocess_input(
            np.float32(images),
            mode='tf'
        )

    def compile(self):
        self.model.compile(
            loss=loss,
            optimizer=optimizers.Adam()
        )


from typing import Union, Tuple, List
import matplotlib.path as mplPath

import cv2
import numpy as np
import imgaug as ia
from shapely import affinity, geometry
from scipy import spatial

from .image import Image


class Selection:
    """Defines a selection within an image.

    Args:
        points: A list of points defining the selection.
    """
    def __init__(self, points):
        assert len(points) > 1, 'A selection requires at least two points.'
        points = np.float32(points)
        if len(points) == 2:
            x1, y1 = points.min(axis=0)
            x2, y2 = points.max(axis=0)
            points = np.float32([
                [x1, y1],
                [x2, y1],
                [x2, y2],
                [x1, y2]
            ])

        points = cv2.convexHull(points)[:, 0, :]
        self.points = points
        self.bbPath = mplPath.Path(points)

    def __add__(self, other):
        mp1 = geometry.MultiPoint(self.points)
        mp2 = geometry.MultiPoint(other.points)
        x, y = mp1.union(mp2).convex_hull.exterior.xy
        points = np.concatenate([
            np.expand_dims(x, axis=1),
            np.expand_dims(y, axis=1)
        ], axis=1)
        return Selection(points)

    def area(self) -> float:
        """Compute the area of the selection."""
        return cv2.contourArea(
            np.int32(self.points[:, np.newaxis, :])
        )

    def xywh(self) -> Tuple[int, int, int, int]:
        """Get the bounding box as x, y, width
        and height."""
        x1, y1, x2, y2 = self.bbox()
        return x1, y1, x2 - x1, y2 - y1

    def shrink(self, scale: float) -> 'Selection':
        """Obtain a new shrunk version of the selection.

        Args:
            scale: The amount by which to shrink the
                selection. Values less than 1 cause
                shrinking. Values larger than 1 cause
                expansion.
        """
        shrunk = affinity.scale(
            geometry.Polygon(self.points), xfact=scale, yfact=scale
        )
        return Selection(shrunk.exterior.coords)

    def contains_point(self, other) -> bool:
        """Determine whether this selection contains a
        particular point.

        Args:
            other: An iterable describing a point (x, y)

        Returns:
            A bool, True if `other` is contained within
            selection, False otherwise.
        """
        return self.bbPath.contains_point(other)

    def contains_points(self, other):
        """Determine whether this selection contains a
        series of points.

        Args:
            other: A numpy array of shape (N, 2)

        Returns:
            A numpy array of shape (N, ), describing
            whether or not the point at that
            position was in the selection.
        """
        return self.bbPath.contains_points(other)

    def bbox(self) -> Tuple[int, int, int, int]:
        """The simple bounding box containing the selection.

        Returns:
            The coordinates (x1, y1, x2, y2). The first set always
            correspond with the top left of the image. The second
            set always correspond with the bottom right of the image.
        """
        x1, y1 = self.points.min(axis=0)
        x2, y2 = self.points.max(axis=0)
        return [int(v) for v in [x1, y1, x2, y2]]

    def keypoints(self) -> List[ia.Keypoint]:
        """Obtain a list of keypoints for the selection.

        Returns:
            A list of keypoints
        """
        return [ia.Keypoint(x=x, y=y) for x, y in self.points]

    def draw(
        self,
        image: Image,
        color: Union[
            Tuple[int, int, int],
            Tuple[int, int, int, int]
        ],
        opaque: bool=False
    ) -> Image:
        """Draw selection onto given image.

        Args:
            image: The image to draw on.
            color: The color to use.
            opaque: Whether the box should be filled.

        Returns:
            The image with the selection drawn
        """
        target = image.copy()
        pts = np.array(
            [tuple(map(int, (x, y))) for x, y in self.points]
        )
        if opaque:
            return cv2.fillPoly(
                img=target,
                pts=[pts],
                color=color
            )
        else:
            return cv2.polylines(
                img=target,
                pts=[pts],
                isClosed=True,
                thickness=5,
                color=color
            )

    def extract(self, image):
        """Extract selection from image (i.e., crop the image
        to the selection).
        """
        x1, y1, x2, y2 = self.bbox()
        cropped = image[max(y1, 0):max(y2, 0), max(x1, 0):max(0, x2)]
        return cropped

    def assign_keypoints(self, keypoints: ia.KeypointsOnImage) -> 'Selection':
        """Obtain a revised version of the selection
        with the given keypoints"""
        if len(keypoints) != len(self.keypoints()):
            raise ValueError(
                'The wrong number of keypoints were provided.'
            )
        points = [
            (kp.x, kp.y) for kp in keypoints
        ]
        return Selection(
            points=points
        )

    def crop(self, width, height):
        """Crop a selection to a given image width
        and height.

        Args:
            width: The width of the image
            height: The height of the image
        """
        pts = self.points.copy()
        pts[:, 0] = pts[:, 0].clip(0, width)
        pts[:, 1] = pts[:, 1].clip(0, height)
        return Selection(points=pts)

    def resize(self, scale: float) -> 'Selection':
        """Obtain a revised selection with a given
        uniform scaling."""
        return Selection(
            points=[
                (scale*x, scale*y) for x, y in self.points
            ]
        )

    def rbox(self) -> Tuple[
            Tuple[float, float],
            Tuple[float, float],
            Tuple[float, float],
            Tuple[float, float],
            float
    ]:
        """Obtain the parameters of a rotated box.

        Returns:
            The vertices of the rotated box in top-left,
            top-right, bottom-right, bottom-left order along
            with the angle of rotation about the bottom left corner.
        """
        mp = geometry.MultiPoint(points=self.points)
        pts = np.array(list(zip(*mp.minimum_rotated_rectangle.exterior.xy)))[:-1]  # noqa: E501

        # The code below is taken from
        # https://github.com/jrosebr1/imutils/blob/master/imutils/perspective.py

        # sort the points based on their x-coordinates
        xSorted = pts[np.argsort(pts[:, 0]), :]

        # grab the left-most and right-most points from the sorted
        # x-roodinate points
        leftMost = xSorted[:2, :]
        rightMost = xSorted[2:, :]

        # now, sort the left-most coordinates according to their
        # y-coordinates so we can grab the top-left and bottom-left
        # points, respectively
        leftMost = leftMost[np.argsort(leftMost[:, 1]), :]
        (tl, bl) = leftMost

        # now that we have the top-left coordinate, use it as an
        # anchor to calculate the Euclidean distance between the
        # top-left and right-most points; by the Pythagorean
        # theorem, the point with the largest distance will be
        # our bottom-right point
        D = spatial.distance.cdist(tl[np.newaxis], rightMost, "euclidean")[0]
        (br, tr) = rightMost[np.argsort(D)[::-1], :]

        # return the coordinates in top-left, top-right,
        # bottom-right, and bottom-left order
        pts = np.array([tl, tr, br, bl], dtype="float32")

        rotation = np.arctan((tl[0] - bl[0]) / (tl[1] - bl[1]))
        return pts, rotation


from typing import Union, Tuple
from os import path
import logging
import urllib
import io

import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
import validators
import cv2

log = logging.getLogger(__name__)


class Image(np.ndarray):
    """Provides convenience functions on top of `np.ndarray`
    for image handling.
    """
    def __init__(self, *args, **kwargs):
        return super(Image, self).__init__(*args, **kwargs)

    @staticmethod
    def read(filepath_or_buffer: Union[str, io.BytesIO]):
        """Read a file into an image object

        Args:
            filepath_or_buffer: The path to the file or any object
                with a `read` method (such as `io.BytesIO`)
        """
        if hasattr(filepath_or_buffer, 'read'):
            image = np.asarray(
                bytearray(filepath_or_buffer.read()),
                dtype=np.uint8
            )
            image = cv2.imdecode(image, cv2.IMREAD_UNCHANGED)
        elif (
            type(filepath_or_buffer) == str and
            validators.url(filepath_or_buffer)
        ):
            return Image.read(urllib.request.urlopen(filepath_or_buffer))
        else:
            assert path.isfile(filepath_or_buffer), \
                'Could not find image at path: ' + filepath_or_buffer
            image = cv2.imread(filepath_or_buffer)
        return image.view(Image).rbswap()

    @staticmethod
    def new(
        width: int,
        height: int,
        channels: int,
        cval=255
    ) -> 'Image':
        """Obtain a new blank image with given dimensions.

        Args:
            width: The width of the blank image
            height: The height of the blank image
            channels: The number of channels. If 0, the image
                will only have two dimensions (y and x).
            cval: The value to set all pixels to (does not apply to the
                alpha channel)

        Returns:
            The blank image
        """
        if channels == 0:
            image = np.zeros((height, width)) + 255
        else:
            image = np.zeros((height, width, channels)) + cval
        return np.uint8(image).view(Image)

    @property
    def width(self):
        """The width of the image"""
        return self.shape[1]

    @property
    def height(self):
        """The height of the image"""
        return self.shape[0]

    @property
    def channels(self):
        """The number of channels in the image"""
        return 0 if len(self.shape) == 2 else self.shape[2]

    def rbswap(self):
        """Swap the red and blue channels for reading and writing
        images."""
        if self.channels != 3 and self.channels != 4:
            log.info('Not swapping red and blue due to number of channels.')
            return self
        return cv2.cvtColor(self, cv2.COLOR_RGB2BGR).view(Image)

    def color(self, channels=3):
        """Convert to color image if it is not already."""
        if len(self.shape) == 3 and self.shape[2] == 1:
            return self.repeat(channels, axis=2)
        elif len(self.shape) == 2:
            image = self[:, :, np.newaxis]
            return image.repeat(channels, axis=2)
        else:
            return self

    def scaled(self, minimum: float=-1, maximum: float=1):
        """Obtain a scaled version of the image with values between
        minimum and maximum.

        Args:
            minimum: The minimum value
            maximum: The maximum value

        Returns:
            An array of same shape as image but of dtype `np.float32` with
            values scaled appropriately.
        """
        assert maximum > minimum
        x = np.float32(self)
        x /= 255
        x *= (maximum - minimum)
        x += minimum
        return x

    def fit(
        self,
        width: int=None,
        height: int=None,
        cval: int=255
    ) -> Tuple['Image', float]:
        """Obtain a new image, fit to the specified size.

        Args:
            width: The new width
            height: The new height
            cval: The constant value to use to fill the remaining areas of
                the image

        Returns:
            The new image and the scaling that was applied.
        """
        if width == self.width and height == self.height:
            return self, 1
        scale = min(width / self.width, height / self.height)
        fitted = Image.new(
            width=width,
            height=height,
            channels=self.channels,
            cval=cval
        ).view(Image)
        image = self.resize(scale=scale)
        fitted[:image.height, :image.width] = image[:width, :height]
        return fitted, scale

    def resize(self, scale: float, interpolation=cv2.INTER_NEAREST):
        """Obtain resized version of image with a given scale

        Args:
            scale: The scale by which to resize the image
            interpolation: The interpolation method to use

        Returns:
            The scaled image
        """
        width = int(np.ceil(scale*self.width))
        height = int(np.ceil(scale*self.height))
        resized = cv2.resize(
            self,
            dsize=(width, height),
            interpolation=interpolation
        ).view(Image)
        if len(resized.shape) == 2 and len(self.shape) == 3:
            # This was a grayscale image and we need it to be returned
            # as such.
            resized = resized[:, :, np.newaxis]
        return resized

    def buffer(self, extension='.jpg') -> io.BytesIO:
        """Convert the image to a BytesIO object with encoding
        for the given extension.

        Args:
            extension: The extension of the target file format,
                such as '.jpg'

        Returns:
            The encoded file
        """
        image = cv2.imencode(extension, self.rbswap())[1].tobytes()
        return io.BytesIO(image)

    def save(
        self,
        filepath_or_buffer: Union[str, io.BytesIO],
        extension='.jpg'
    ):
        """Save the image

        Args:
            filepath_or_buffer: The file or buffer to
                which to save the image. If buffer,
                format must be provided
            esxtension: The extension for the format to use
                if writing to buffer
        """
        image = self.rbswap()
        if hasattr(filepath_or_buffer, 'write'):
            data = cv2.imencode(extension, image)[1].tobytes()
            filepath_or_buffer.write(data)
        else:
            cv2.imwrite(filepath_or_buffer, img=image)

    def show(self, ax: mpl.axes.Axes=None) -> mpl.axes.Axes:
        """Show an image

        Args:
            ax: Axis on which to show the image

        Returns:
            An axes object
        """
        if ax is None:
            ax = plt
        if len(self.shape) == 3 and self.shape[2] >= 3:
            return ax.imshow(self)
        elif len(self.shape) == 3 and self.shape[2] == 1:
            return ax.imshow(self[:, :, 0])
        elif len(self.shape) == 2:
            return ax.imshow(self)
        else:
            raise ValueError('Incorrect dimensions for image data.')
        return ax.imshow(self)


import io
import logging
import configparser
from collections import defaultdict
from functools import reduce

import numpy as np
from keras import backend as K
from keras import layers as keras_layers
from keras.models import Model
from keras.regularizers import l2

log = logging.getLogger(__name__)


def compose(*funcs):
    """Compose arbitrarily many functions, evaluated left to right.

    Reference: https://mathieularose.com/function-composition-in-python/
    """
    if funcs:
        return reduce(lambda f, g: lambda *a, **kw: g(f(*a, **kw)), funcs)
    else:
        raise ValueError('Composition of empty sequence not supported.')


def convert_darknet_weights(
    config_path: str,
    weights_path: str,
    output_path: str,
    weights_only: bool=False
):
    """Convert DarkNet weights to Keras weights. This function
    taken directly from:
    https://github.com/qqwweee/keras-yolo3/blob/master/convert.py

    """
    def unique_config_sections(config_file):
        """Convert all config sections to have unique names. Adds unique suffixes
        to config sections for compability with configparser.
        """
        section_counters = defaultdict(int)
        output_stream = io.StringIO()
        with open(config_file) as fin:
            for line in fin:
                if line.startswith('['):
                    section = line.strip().strip('[]')
                    _section = section + '_' + str(section_counters[section])
                    section_counters[section] += 1
                    line = line.replace(section, _section)
                output_stream.write(line)
        output_stream.seek(0)
        return output_stream

    assert config_path.endswith('.cfg'), '{} is not a .cfg file'.format(config_path)  # noqa: E501
    assert weights_path.endswith('.weights'), '{} is not a .weights file'.format(weights_path)  # noqa: E501
    assert output_path.endswith('.h5'), 'output path {} is not a .h5 file'.format(output_path)  # noqa: E501

    log.info('Loading weights.')
    weights_file = open(weights_path, 'rb')
    major, minor, revision = np.ndarray(
        shape=(3, ), dtype='int32', buffer=weights_file.read(12))
    if (major*10+minor) >= 2 and major < 1000 and minor < 1000:
        seen = np.ndarray(
            shape=(1,),
            dtype='int64',
            buffer=weights_file.read(8)
        )
    else:
        seen = np.ndarray(
            shape=(1,),
            dtype='int32',
            buffer=weights_file.read(4)
        )
    log.info(
        'Weights header: {0}, {1}, {2}, {3}'.format(
            major,
            minor,
            revision,
            seen
        )
    )

    log.info('Parsing Darknet config.')
    unique_config_file = unique_config_sections(config_path)
    cfg_parser = configparser.ConfigParser()
    cfg_parser.read_file(unique_config_file)

    log.info('Creating Keras model.')
    index = {
        'input': 0,
        'zeropadding2d': 0,
        'convolution2d': 0,
        'concatenate': 0,
        'maxpooling2d': 0,
        'add': 0,
        'upsampling2d': 0,
        'batchnormalization': 0,
        'leakyrelu': 0
    }
    input_layer = keras_layers.Input(
        shape=(None, None, 3),
        name='input_{0}'.format(index['input'])
    )
    index['input'] += 1
    prev_layer = input_layer
    all_layers = []

    if 'net_0' in cfg_parser.sections():
        weight_decay = float(cfg_parser['net_0']['decay'])
    else:
        weight_decay = 5e-4

    count = 0
    out_index = []
    for section in cfg_parser.sections():
        log.info('Parsing section {}'.format(section))
        if section.startswith('convolutional'):
            filters = int(cfg_parser[section]['filters'])
            size = int(cfg_parser[section]['size'])
            stride = int(cfg_parser[section]['stride'])
            pad = int(cfg_parser[section]['pad'])
            activation = cfg_parser[section]['activation']
            batch_normalize = 'batch_normalize' in cfg_parser[section]

            padding = 'same' if pad == 1 and stride == 1 else 'valid'

            # Setting weights.
            # Darknet serializes convolutional weights as:
            # [bias/beta, [gamma, mean, variance], conv_weights]
            prev_layer_shape = K.int_shape(prev_layer)

            weights_shape = (size, size, prev_layer_shape[-1], filters)
            darknet_w_shape = (filters, weights_shape[2], size, size)
            weights_size = np.product(weights_shape)
            log.info(
                'conv2d (activation: {0}, weights: {1}, bn: {2})'.format(
                    str(activation),
                    str(weights_shape),
                    str(batch_normalize)
                )
            )
            conv_bias = np.ndarray(
                shape=(filters, ),
                dtype='float32',
                buffer=weights_file.read(filters * 4))
            count += filters

            if batch_normalize:
                bn_weights = np.ndarray(
                    shape=(3, filters),
                    dtype='float32',
                    buffer=weights_file.read(filters * 12))
                count += 3 * filters

                bn_weight_list = [
                    bn_weights[0],  # scale gamma
                    conv_bias,  # shift beta
                    bn_weights[1],  # running mean
                    bn_weights[2]  # running var
                ]

            conv_weights = np.ndarray(
                shape=darknet_w_shape,
                dtype='float32',
                buffer=weights_file.read(weights_size * 4))
            count += weights_size

            # DarkNet conv_weights are serialized Caffe-style:
            # (out_dim, in_dim, height, width)
            # We would like to set these to Tensorflow order:
            # (height, width, in_dim, out_dim)
            conv_weights = np.transpose(conv_weights, [2, 3, 1, 0])
            conv_weights = [conv_weights] if batch_normalize else [
                conv_weights, conv_bias
            ]

            # Handle activation.
            act_fn = None
            if activation == 'leaky':
                pass  # Add advanced activation later.
            elif activation != 'linear':
                raise ValueError(
                    'Unknown activation function `{}` in section {}'.format(
                        activation, section))

            # Create Conv2D layer
            if stride > 1:
                # Darknet uses left and top padding instead of 'same' mode
                prev_layer = keras_layers.ZeroPadding2D(
                    ((1, 0), (1, 0)),
                    name='zeropadding2d_{0}'.format(index['zeropadding2d'])
                )(prev_layer)
                index['zeropadding2d'] += 1
            conv_layer = keras_layers.Conv2D(
                filters=filters,
                kernel_size=(size, size),
                strides=(stride, stride),
                kernel_regularizer=l2(weight_decay),
                use_bias=not batch_normalize,
                weights=conv_weights,
                activation=act_fn,
                padding=padding,
                name='convolution2d_{0}'.format(index['convolution2d'])
            )(prev_layer)
            index['convolution2d'] += 1

            if batch_normalize:
                conv_layer = keras_layers.BatchNormalization(
                    weights=bn_weight_list,
                    name='batchnormalization_{0}'.format(
                        index['batchnormalization']
                    )
                )(conv_layer)
                index['batchnormalization'] += 1
            prev_layer = conv_layer

            if activation == 'linear':
                all_layers.append(prev_layer)
            elif activation == 'leaky':
                act_layer = keras_layers.LeakyReLU(
                    alpha=0.1,
                    name='leakyrelu_{0}'.format(index['leakyrelu'])
                )(prev_layer)
                index['leakyrelu'] += 1
                prev_layer = act_layer
                all_layers.append(act_layer)

        elif section.startswith('route'):
            ids = [int(i) for i in cfg_parser[section]['layers'].split(',')]
            layers = [all_layers[i] for i in ids]
            if len(layers) > 1:
                log.info(
                    'Concatenating route layers: ' + ','.join(
                        [str(l) for l in layers]
                    )
                )
                concatenate_layer = keras_layers.Concatenate(
                    name='concatenate_{0}'.format(index['concatenate'])
                )(layers)
                index['concatenate'] += 1
                all_layers.append(concatenate_layer)
                prev_layer = concatenate_layer
            else:
                skip_layer = layers[0]  # only one layer to route
                all_layers.append(skip_layer)
                prev_layer = skip_layer

        elif section.startswith('maxpool'):
            size = int(cfg_parser[section]['size'])
            stride = int(cfg_parser[section]['stride'])
            all_layers.append(
                keras_layers.MaxPooling2D(
                    pool_size=(size, size),
                    strides=(stride, stride),
                    padding='same',
                    name='maxpooling2d_{0}'.format(index['maxpooling2d'])
                )(prev_layer)
            )
            index['maxpooling2d'] += 1
            prev_layer = all_layers[-1]

        elif section.startswith('shortcut'):
            shortcut_index = int(cfg_parser[section]['from'])
            activation = cfg_parser[section]['activation']
            assert activation == 'linear', 'Only linear activation supported.'
            all_layers.append(keras_layers.Add(
                name='add_{0}'.format(index['add'])
            )([all_layers[shortcut_index], prev_layer]))
            index['add'] += 1
            prev_layer = all_layers[-1]

        elif section.startswith('upsample'):
            stride = int(cfg_parser[section]['stride'])
            assert stride == 2, 'Only stride=2 supported.'
            all_layers.append(
                keras_layers.UpSampling2D(
                    stride,
                    name='upsampling2d_{0}'.format(index['upsampling2d'])
                )(prev_layer)
            )
            index['upsampling2d'] += 1
            prev_layer = all_layers[-1]

        elif section.startswith('yolo'):
            out_index.append(len(all_layers)-1)
            all_layers.append(None)
            prev_layer = all_layers[-1]

        elif section.startswith('net'):
            pass

        else:
            raise ValueError(
                'Unsupported section header type: {}'.format(section))

    # Create and save model.
    if len(out_index) == 0:
        out_index.append(len(all_layers)-1)
    model = Model(
        inputs=input_layer,
        outputs=[all_layers[i] for i in out_index]
    )
    if weights_only:
        model.save_weights('{}'.format(output_path))
        log.info('Saved weights to {0}'.format(output_path))
    else:
        model.save('{}'.format(output_path))
        log.info('Saved model to {0}'.format(output_path))

    # Check to see if all weights have been read.
    remaining_weights = len(weights_file.read()) / 4
    weights_file.close()
    log.info(
        'Read {} of {} from Darknet weights'.format(
            count,
            count + remaining_weights
        )
    )
    if remaining_weights > 0:
        log.warning('{0} unused weights'.format(remaining_weights))
    return model



from abc import ABC, abstractmethod
from typing import List, Union, Tuple

from imgaug import augmenters as iaa
import numpy as np

from .. import metrics
from ..core import (SceneCollection, Annotation, Image)


class Detector(ABC):
    model = None

    def detect(self, image: Union[Image, np.ndarray],
               threshold: float = 0.5) -> List[Annotation]:
        """Run detection for a given image.

        Args:
            image: The image to run detection on

        Returns:
            A list of annotations
        """
        image = image.view(Image)
        image, scale = self._scale_to_model_size(image)
        X = self.compute_inputs([image])
        y = self.model.predict(X)
        annotations = self.invert_targets(
            y, images=[image], threshold=threshold)[0].annotations
        return [a.resize(1 / scale) for a in annotations]

    def detect_batch(self,
                     images: List[Union[Image, np.ndarray]],
                     threshold: float = 0.5,
                     batch_size: int = 32) -> List[List[Annotation]]:
        """
        Perform object detection on a batch of images.

        Args:
            images: A list of images
            threshold: The detection threshold for the images
            batch_size: The batch size to use with the underlying model

        Returns:
            A list of lists of annotations.
        """
        images = [image.view(Image) for image in images]
        images, scales = list(
            zip(*[self._scale_to_model_size(image) for image in images]))
        X = self.compute_inputs(images)
        y = self.model.predict(X, batch_size=batch_size)
        scenes = self.invert_targets(y, images=images, threshold=threshold)
        return [[a.resize(1 / scale) for a in scene.annotations]
                for scene, scale in zip(scenes, scales)]

    def _scale_to_model_size(self, image: Image):
        height, width = self.model.input_shape[1:3]
        if height is not None and width is not None:
            image, scale = image.fit(width=width, height=height)
        else:
            scale = 1
        return image, scale

    @abstractmethod
    def compute_inputs(self, images: List[Image]) -> np.ndarray:
        """Convert images into suitable model inputs. *You
        usually should not need this method*. For training,
        use `detector.train()`. For detection, use
        `detector.detect()`.

        Args:
            images: The images to convert

        Returns:
            The input to the model
        """
        pass

    @abstractmethod
    def compute_targets(self, collection: SceneCollection
                        ) -> Union[List[np.ndarray], np.ndarray]:
        """Compute the expected outputs for a model. *You
        usually should not need this method*. For training,
        use `detector.train()`. For detection, use
        `detector.detect()`.

        Args:
            collection: The scene collection for which the
                outputs should be calculated.

        Returns:
            The output(s) as a list of numpy arrays
        """
        pass

    def freeze_backbone(self):
        """Freeze the body of the model, leaving the final classification and
        regression layer as trainable."""
        for l in self.backbone.layers:
            l.trainable = False
        self.compile()

    def unfreeze_backbone(self):
        """Unfreeze the body of the model, making all layers trainable."""
        for l in self.backbone.layers:
            l.trainable = True
        self.compile()

    def compute_Xy(self, collection: SceneCollection):
        return (self.compute_inputs(collection.images),
                self.compute_targets(collection))

    def batch_generator(self,
                        collection: SceneCollection,
                        train_shape: Tuple[int, int, int],
                        batch_size: int = 1,
                        augmenter: iaa.Augmenter = None):
        while True:
            sample = collection.sample(n=batch_size).augment(
                augmenter=augmenter).fit(
                    height=train_shape[0], width=train_shape[1])[0]
            yield self.compute_Xy(sample)

    def train(self,
              training: SceneCollection,
              validation: SceneCollection = None,
              batch_size: int = 1,
              augmenter: iaa.Augmenter = None,
              train_shape: Tuple[int, int, int] = None,
              **kwargs):
        """Run training job. All additional keyword arguments
        passed to Keras' `fit_generator`. Note that, at a minimum,
        you must provide the `epochs` and `steps_per_epoch` arguments.
        If validation data is provided, you will also need to provide a
        `validation_steps` argument.

        Args:
            training: The collection of training images
            validation: The collection of validation images
            batch_size: The batch size to use for training
            augmenter: The augmenter for generating samples
        """
        assert training.annotation_config == self.annotation_config, \
            'The training set configuration clashes with detector'
        assert (
            validation is None or
            validation.annotation_config == self.annotation_config
        ), \
            'The validation set configuration clashes with detector'
        assert training.consistent, \
            'The training set has inconsistent annotation configuration'
        assert (
            validation is None or
            validation.consistent
        ), \
            'The validation set has inconsistent annotation configuration'
        if hasattr(self, 'training_model'):
            training_model = self.training_model
        else:
            training_model = self.model
        if train_shape is None:
            train_shape = training_model.input_shape[1:]
            assert all(s is not None for s in train_shape), \
                'train_shape must be provided for this model.'
        training_generator = self.batch_generator(
            collection=training,
            batch_size=batch_size,
            augmenter=augmenter,
            train_shape=train_shape)
        if validation is None:
            history = training_model.fit_generator(
                generator=training_generator, **kwargs)
        else:
            validation = validation.fit(
                height=train_shape[0], width=train_shape[1])[0]
            history = training_model.fit_generator(
                generator=training_generator,
                validation_data=self.compute_Xy(validation),
                **kwargs)
        return history

    def mAP(self, collection: SceneCollection, iou_threshold=0.5):
        """Compute the mAP metric for a given collection
        of ground truth scenes.

        Args:
            collection: The collection to evaluate
            iou_threshold: The IoU threshold required for
                a match

        Returns:
            mAP score
        """
        pred = collection.assign(scenes=[
            s.assign(annotations=self.detect(s.image, threshold=0))
            for s in collection
        ])
        return metrics.mAP(
            true_collection=collection,
            pred_collection=pred,
            iou_threshold=iou_threshold)



from typing import Union, List, Tuple
from os import path
import numpy as np
import logging
import math
import io

from sklearn.model_selection import train_test_split
from imgaug import augmenters as iaa
import matplotlib.pyplot as plt
import matplotlib as mpl
import imgaug as ia
import validators

from .annotation import AnnotationConfiguration, Annotation
from .image import Image

log = logging.getLogger(__name__)


class Scene:
    """A single annotated image.

    Args:
        annotation_config: The configuration for annotations for the
            image.
        annotations: The list of annotations.
        image: The image that was annotated. Can be lazy-loaded by passing
            a string filepath.
        cache: Defines caching behavior for the image. If `True`, image is
            loaded into memory the first time that the image is requested.
            If `False`, image is loaded from the file path or URL whenever
            the image is requested.
    """

    def __init__(self,
                 annotation_config: AnnotationConfiguration,
                 annotations: List[Annotation],
                 image: Union[Image, np.ndarray, str],
                 cache: bool = True):
        assert isinstance(image, np.ndarray) or isinstance(image, str), \
            'Image must be string or ndarray, not ' + str(type(image))
        if isinstance(image, np.ndarray):
            image = image.view(Image)
        self._image = image
        self._annotations = annotations
        self._annotation_config = annotation_config
        self.cache = cache

    @property
    def image(self) -> Image:
        """The image that is being annotated"""
        # Check to see if we have an actual image
        # or just a string
        if type(self._image) != str:
            return self._image

        # Check the cache first if image is a URL
        if (validators.url(self._image) and type(self.cache) == str
                and path.isfile(self.cache)):
            self._image = self.cache

        # Load the image
        log.debug('Reading from ' + self._image)
        image = Image.read(self._image)

        # Check how to handle caching the image
        # for future reads
        if self.cache is True:
            self._image = image
        elif self.cache is False:
            pass
        else:
            raise ValueError('Cannot handle cache parameter: {0}.'.format(
                self.cache))
        return image

    @property
    def annotation_config(self):
        """The annotation configuration"""
        return self._annotation_config

    @property
    def annotations(self) -> List[Annotation]:
        """Get the list of annotations"""
        return self._annotations

    def assign(self, **kwargs) -> 'Scene':
        """Get a new scene with only the supplied
        keyword arguments changed."""
        if 'annotation_config' in kwargs:
            # We need to change all the categories for annotations
            # to match the new annotation configuration.
            annotations = kwargs.get('annotations', self.annotations)
            annotation_config = kwargs['annotation_config']
            revised = [
                ann.convert(annotation_config=annotation_config)
                for ann in annotations
            ]
            revised = [ann for ann in revised if ann is not None]
            removed = len(annotations) - len(revised)
            log.debug('Removed {0} annotations when changing '.format(removed)
                      + 'annotation configuration.')
            kwargs['annotations'] = revised
        # We use the _image instead of image to avoid triggering an
        # unnecessary read of the actual image.
        defaults = {
            'annotation_config': self.annotation_config,
            'annotations': self.annotations,
            'image': self._image,
            'cache': self.cache
        }
        kwargs = {**defaults, **kwargs}
        return Scene(**kwargs)

    def show(self, *args, **kwargs) -> mpl.axes.Axes:
        """Show an annotated version of the image. All arguments
        passed to `Image.show()`.
        """
        return self.annotated().show(*args, **kwargs)

    def scores(self):
        """Obtain an array containing the confidence
        score for each annotation."""
        return np.array([a.score for a in self.annotations])

    def bboxes(self):
        """Obtain an array of shape (N, 5) where the columns are
        x1, y1, x2, y2, class_index where class_index is determined
        from the annotation configuration."""
        return np.array([
            a.selection.bbox() + [self.annotation_config.index(a.category)
                                  ]  # noqa: E501
            for a in self.annotations
        ])

    def fit(self, width, height):
        """Obtain a new scene fitted to the given width and height.

        Args:
            width: The new width
            height: The new height

        Returns:
            The new scene and the scale
        """
        image, scale = self.image.fit(width=width, height=height)
        annotations = [ann.resize(scale=scale) for ann in self.annotations]
        return self.assign(image=image, annotations=annotations), scale

    def annotated(self,
                  dpi=72,
                  fontsize='x-large',
                  labels=True,
                  opaque=False,
                  color=(255, 0, 0)) -> Image:
        """Show annotations on the image itself.

        Args:
            dpi: The resolution for the image
            fontsize: How large to show labels
            labels: Whether or not to show labels
            opaque: Whether to draw annotations filled
                in.
            color: The color to use for annotations.
        """
        plt.ioff()
        fig, ax = plt.subplots()
        ax.set_xlabel('')
        ax.set_xticks([])
        ax.set_ylabel('')
        ax.set_yticks([])
        ax.axis('off')
        img_raw = self.image
        img = img_raw
        for ann in self.annotations:
            img = ann.selection.draw(img, color=color, opaque=opaque)
        img.show(ax=ax)
        if labels:
            for ann in self.annotations:
                x1, y1, x2, y2 = ann.selection.bbox()
                ax.annotate(
                    s=ann.category.name,
                    xy=(x1, y1),
                    fontsize=fontsize,
                    backgroundcolor=(1, 1, 1, 0.5))
        ax.set_xlim(0, img_raw.width)
        ax.set_ylim(img_raw.height, 0)
        fig.canvas.draw()
        raw = io.BytesIO()
        fig.savefig(
            raw,
            dpi=dpi,
            frameon=True,
            pad_inches=0,
            transparent=False,
            bbox_inches='tight')
        plt.close(fig)
        plt.ion()
        raw.seek(0)
        img = Image.read(raw)
        img = img[:, :, :3]
        raw.close()
        return img

    def resize(self, scale) -> 'Scene':
        """Obtain a resized version of the scene.

        Args:
            scale: The scale for the new scene.
        """
        return self.assign(
            image=self.image.resize(scale),
            annotations=[ann.resize(scale) for ann in self.annotations])

    def augment(self, augmenter: iaa.Augmenter = None,
                threshold: float = 0.25) -> 'Scene':
        """Obtain an augmented version of the scene using the given augmenter.

        Returns:
            The augmented scene
        """
        if augmenter is None:
            return self
        aug = augmenter.to_deterministic()
        keypoints = []
        keypoints_map = {}
        for i, ann in enumerate(self.annotations):
            current = ann.selection.keypoints()
            startIdx = len(keypoints)
            endIdx = startIdx + len(current)
            keypoints_map[i] = (startIdx, endIdx)
            keypoints.extend(current)
        keypoints = ia.KeypointsOnImage(keypoints, shape=self.image.shape)
        image = aug.augment_images([self.image])[0].view(Image)
        keypoints = aug.augment_keypoints([keypoints])[0].keypoints
        annotations = []
        for i, ann in enumerate(self.annotations):
            startIdx, endIdx = keypoints_map[i]
            current = keypoints[startIdx:endIdx]
            selection = ann.selection.assign_keypoints(current)
            area = selection.area()
            selection = selection.crop(width=image.width, height=image.height)
            if area == 0 or (selection.area() / area < threshold):
                continue
            annotations.append(ann.assign(selection=selection))
        return self.assign(image=image, annotations=annotations)


class SceneCollection:
    """A collection of scenes.

    Args:
        annotation_config: The configuration that should be used for all
            underlying scenes.
        scenes: The list of scenes.
    """

    def __init__(self,
                 scenes: List[Scene],
                 annotation_config: AnnotationConfiguration = None):
        assert len(scenes) > 0, \
            'A scene collection must have at least one scene'
        if annotation_config is None:
            annotation_config = scenes[0].annotation_config
        for i, s in enumerate(scenes):
            if s.annotation_config != annotation_config:
                raise ValueError(
                    'Scene {0} of {1} has inconsistent configuration.'.format(
                        i + 1, len(scenes)))
        self._annotation_config = annotation_config
        self._scenes = scenes

    def __getitem__(self, key):
        return self.scenes[key]

    def __len__(self):
        return len(self._scenes)

    @property
    def scenes(self):
        """The list of scenes"""
        return self._scenes

    @property
    def annotation_config(self):
        """The annotation configuration"""
        return self._annotation_config

    @property
    def uniform(self):
        """Specifies whether all scenes in the collection are
        of the same size. Note: This will trigger an image load."""
        return np.unique(
            np.array([s.image.shape for s in self.scenes]),
            axis=0).shape[0] == 1  # noqa: E501

    @property
    def consistent(self):
        """Specifies whether all scenes have the same annotation
        configuration."""
        return all(s.annotation_config == self.annotation_config
                   for s in self.scenes)  # noqa: E501

    @property
    def images(self):
        return [s.image for s in self.scenes]

    def augment(self, **kwargs):
        """Obtained an augmented version of the given collection.
        All arguments passed to `Scene.augment`"""
        return self.assign(scenes=[s.augment(**kwargs) for s in self.scenes])

    def train_test_split(self, *args, **kwargs
                         ) -> Tuple['SceneCollection', 'SceneCollection']:
        """Obtain new scene collections, split into train
        and test. All arguments passed to
        `sklearn.model_selection.train_test_split
        <https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.train_test_split.html>`_.

        For example, to get two collections, one
        containing 70% of the data and the other containing
        30%, ::

            training, testing = collection.train_test_split(
                train_size=0.7, test_size=0.3
            )

        You can also use the `stratify` argument to ensure an even split
        between different kinds of scenes. For example, to split
        scenes containing at least 3 annotations proportionally, ::

            training, testing = collection.train_test_split(
                train_size=0.7, test_size=0.3,
                stratify=[len(s.annotations) >= 3 for s in collection]
            )


        Returns:
            A train and test scene collection.
        """
        train, test = train_test_split(self.scenes, *args, **kwargs)
        return (self.assign(scenes=train), self.assign(scenes=test))

    def assign(self, **kwargs) -> 'Scene':
        """Obtain a new scene with the given keyword arguments
        changing. If `annotation_config` is provided, the annotations
        are converted to the new `annotation_config` first.

        Returns:
            A new scene

        """
        if 'annotation_config' in kwargs:
            annotation_config = kwargs['annotation_config']
            scenes = kwargs.get('scenes', self.scenes)
            kwargs['scenes'] = [
                s.assign(annotation_config=annotation_config) for s in scenes
            ]
        defaults = {
            'scenes': self.scenes,
            'annotation_config': self.annotation_config
        }
        kwargs = {**defaults, **kwargs}
        return SceneCollection(**kwargs)

    def sample(self, n, replace=True) -> 'SceneCollection':
        """Get a random subsample of this collection"""
        selected = np.random.choice(len(self.scenes), n, replace=replace)
        return self.assign(scenes=[self.scenes[i] for i in selected])

    def fit(self, **kwargs):
        """Obtain a new scene collection, fitted to the given width
        and height. All arguments passed to `scene.fit()`

        Returns:
            The new scene collection and list of scales.
        """
        scales = []
        scenes = []
        for s in self.scenes:
            scene, scale = s.fit(**kwargs)
            scenes.append(scene)
            scales.append(scale)
        return self.assign(scenes=scenes), scales

    def thumbnails(self, n=10, width=200, height=200, ncols=2):
        """Get a thumbnail sample of the images in the collection.

        Args:
            n: The number of images to include
            width: The width of each thumbnail
            height: The height of each thumbnail
            ncols: The number of columsn in which to arrange the
                thumbnails.

        Returns:
            The thumbnail image with width ncols*width and height
            (n / ncols)*height
        """
        if n > len(self.scenes):
            log.warning('Collection only has {0} scenes but '
                        'you requested {1} thumbnails. Limiting '
                        'to {0} thumbnails.'.format(len(self.scenes), n))
            n = len(self.scenes)
        nrows = math.ceil(n / ncols)
        sample_indices = np.random.choice(len(self.scenes), n, replace=False)
        sample = [self.scenes[i] for i in sample_indices]
        thumbnails = [scene.annotated() for scene in sample]
        thumbnails = [t.fit(width=width, height=height)[0] for t in thumbnails]
        thumbnail = Image.new(
            width=ncols * width, height=nrows * height, channels=3)
        for rowIdx in range(nrows):
            for colIdx in range(ncols):
                if len(thumbnails) == 0:
                    break
                thumbnail[rowIdx * width:(rowIdx + 1) * width, colIdx *
                          height:(colIdx + 1) * height] = thumbnails.pop()
        return thumbnail
